from decimal import Decimal

import numpy as np
from typing import Tuple, List, Optional

from mygrad._utils import reduce_broadcast


def to_decimal_array(arr):
    """ Convert numpy ND-array to Decimal-type object array of the same shape.
        Used for facilitating high-precision arithmetic.

        Parameters
        ----------
        arr : Union[float, numpy.ndarray]

        Returns
        -------
        numpy.ndarray
            Decimal-type object array"""
    arr = np.asarray(arr)

    if arr.dtype.kind == "O":
        return arr
    return np.array(tuple(Decimal(float(i)) for i in arr.flat), dtype=Decimal).reshape(arr.shape)


def numerical_gradient(f, *args, back_grad, vary_ind=None,
                       h=Decimal(1)/Decimal(int(1e8)),
                       as_decimal=True, kwargs=None):
    """ Computes numerical partial derivatives of f(x0, x1, ...) in each
        of its variables, using the central difference method.
        This is a "fast" method - it varies entire arrays at once. Thus
        this is only appropriate for trivial vectorized functions that
        map accross entries of arrays (like add or multiply). E.g.
        matrix multiplication is *not* suited for this style of gradient.

        Parameters
        ----------
        f : Callable[[numpy.ndarray, ...], numpy.ndarray]
            f(x, ...) -> numpy.ndarray
        *args : Tuple[numpy.ndarray, ...]
            The input arguments to be fed to f.

        back_grad : numpy.ndarray
            The gradient being back-propagated to x and y, via f

        vary_ind : Optional[Tuple[int, ...]]
            If `None`, the partials of f with respect to all the inputs are.
            computed. Otherwise you can specify a sequence of the indices
            of the variables whose partials are to be computed
               0 -> w.r.t x only, 1 -> w.r.t y only, etc.

        h : float, optional, (default=Decimal(1E-8))
            Approximating infinitesimal.

        as_decimal : bool, optional (default=True)
            If True, f's arguments are passed as Decimal-type arrays. This
            improves numerical precision, but is not permitted by some functions.

        kwargs : Optional[Dict]

        Returns
        -------
        Tuple[Union[NoneType, numpy.ndarray], ...]
            df/dx0, df/dx1, ... - evaluated at (`x0`, `x1`, ... ).
        """

    if kwargs is None:
        kwargs = {}

    if not args:
        raise ValueError("At least one value must be passed to `args`")

    h = Decimal(h) if as_decimal else float(h)
    two_h = Decimal(2)*h if as_decimal else 2*h

    args = tuple(to_decimal_array(i) if as_decimal else i for i in args)

    grads = [None]*len(args)

    def gen_fwd_diff(i):
        # x1, ..., x_i + h, ..., xn
        return ((var if j != i else var + h) for j, var in enumerate(args))

    def gen_bkwd_diff(i):
        # x1, ..., x_i - h, ..., xn
        return ((var if j != i else var - h) for j, var in enumerate(args))

    for n in range(len(args)):
        if vary_ind is not None and n not in vary_ind:
            continue
        # central difference in variable n
        dvar = (f(*gen_fwd_diff(n), **kwargs) - f(*gen_bkwd_diff(n), **kwargs)) / (two_h)
        grads[n] = reduce_broadcast(back_grad * dvar.astype(float), args[n].shape)

    return grads


def numerical_gradient_full(f, *args, back_grad, as_decimal=True, kwargs=None, vary_ind=None) -> Tuple[np.ndarray, ...]:
    """ Computes numerical partial derivatives of f(x, y, ..., **kwargs), by
        varying each entry of x, y, ... independently producing a gradient
        in each variable.

        Parameters
        ----------
        f : Callable[[numpy.ndarray, ...], numpy.ndarray]
            f(x, ...) -> numpy.ndarray

        *args : numpy.ndarray
            The array(s) to be passed to f

        back_grad : numpy.ndarray
            The gradient being back-propagated to {x}, via f

        as_decimal : bool, optional (default=True)
            If True, f's arguments are passed as Decimal-type arrays. This
            improves numerical precision, but is not permitted by some functions.

        kwargs : Dict[str, Any], optional (default=dict())
            The keyword arguments to be passed to f.

        vary_ind : Optional[Tuple[int, ...]]
            If `None`, the partials of f with respect to all the inputs are.
            computed. Otherwise you can specify a sequence of the indices
            of the variables whose partials are to be computed
               0 -> w.r.t x only, 1 -> w.r.t y only, etc.

        Returns
        -------
        Tuple[numpy.ndarray, ...]
            df/dx, df/dy, ...
            df/dvar will be None if var was excluded via `vary_ind`
        """
    if kwargs is None:
        kwargs = {}

    args = tuple(to_decimal_array(i) if as_decimal else i for i in args)
    grads = [None] * len(args)  # type: List[Optional[np.ndarray]]
    if isinstance(vary_ind, int):
        vary_ind = [vary_ind]

    for n in range(len(args)):
        if vary_ind is not None and n not in vary_ind:
            continue

        def tmp_f(var: np.ndarray) -> np.ndarray:
            return f(*args[:n], var, *args[n+1:], **kwargs)

        grads[n] = _numerical_gradient_full(tmp_f,
                                            x=args[n],
                                            back_grad=back_grad,
                                            as_decimal=as_decimal)
    return tuple(grads)


def _numerical_gradient_full(f, *, x, back_grad, h=Decimal(1)/Decimal(int(1e8)), as_decimal=True):
    """ Computes numerical partial derivatives of f(x), by
        varying each entry of `x` independently.

        Parameters
        ----------
        f : Callable[[numpy.ndarray], numpy.ndarray]
            f(x) -> numpy.ndarray

        x : numpy.ndarray
            An array storing the sequence(s) of values in the array. More than once
            sequence may be designated, according to the `axis` argument of `f`.

        back_grad : numpy.ndarray
            The gradient being back-propagated to {x}, via f

        h : float, optional, (default=Decimal(1E-8))
            Approximating infinitesimal.

        as_decimal : bool, optional (default=True)
            If True, x is passed to f as a Decimal-type array. This
            improves numerical precision, but is not permitted by some functions.

        Returns
        -------
        numpy.ndarray
            df/dx
        """

    grad = np.empty_like(x)
    x = to_decimal_array(x)
    x_orig = np.copy(x)
    h = Decimal(h)
    back_grad = to_decimal_array(back_grad)

    for ind, val in np.ndenumerate(x):
        x_fwd = x
        x_fwd[ind] = x_orig[ind] + h
        f_fwd = to_decimal_array(f(x_fwd) if as_decimal else f(x_fwd.astype(np.float64)))

        x_bkwd = np.copy(x_fwd)
        x_bkwd[ind] = x_orig[ind] - h
        f_bkwd = to_decimal_array(f(x_bkwd) if as_decimal else f(x_bkwd.astype(np.float64)))

        df_dxi = to_decimal_array((f_fwd - f_bkwd) / (Decimal(2) * h))

        dl_dxi = (df_dxi * back_grad)
        grad[ind] = np.float64(dl_dxi.sum() if isinstance(dl_dxi, np.ndarray) else dl_dxi)

        # reset x
        x[ind] = x_orig[ind]
    return grad.astype(np.float64)



from mygrad.tensor_base import Tensor
from mygrad import multiply_sequence, add_sequence
from numpy.testing import assert_allclose


def test_seq_add():
    a = Tensor(3)
    b = Tensor([1, 2, 3])
    c = Tensor([[1, 2, 3], [2, 3, 4]])
    f = add_sequence(a, b, c, constant=False)
    f.sum().backward()

    a1 = Tensor(3)
    b1 = Tensor([1, 2, 3])
    c1 = Tensor([[1, 2, 3], [2, 3, 4]])
    f1 = a1 + b1 + c1
    f1.sum().backward()

    assert_allclose(f.data, f1.data)
    assert_allclose(f.grad, f1.grad)
    assert_allclose(a.grad, a1.grad)
    assert_allclose(b.grad, b1.grad)
    assert_allclose(c.grad, c1.grad)


def test_seq_mult():
    a = Tensor(3.)
    b = Tensor([1., 2., 3.])
    c = Tensor([[1., 2., 3.], [2., 3., 4.]])
    f = multiply_sequence(a, b, c, constant=False)
    f.sum().backward()

    a1 = Tensor(3.)
    b1 = Tensor([1., 2., 3.])
    c1 = Tensor([[1., 2., 3.], [2., 3., 4.]])
    f1 = a1 * b1 * c1
    f1.sum().backward()

    assert_allclose(f.data, f1.data)
    assert_allclose(f.grad, f1.grad)
    assert_allclose(a.grad, a1.grad)
    assert_allclose(b.grad, b1.grad)
    assert_allclose(c.grad, c1.grad)



from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor


class Dummy(Operation):
    def __call__(self, a, b):
        self.variables = (a, b)
        return 1
    

def dummy(a, b, constant=False):
    return Tensor._op(Dummy, a, b, constant=constant)


def test_constant_arg():
    """ test that the `constant` arg works as intended in Tensor._op"""
    a = Tensor(1)
    b = Tensor(1)
    o_true = dummy(a, b, constant=True)
    assert o_true.constant is True
    assert a._ops == set()
    assert b._ops == set()

    o_false = dummy(a, b, constant=False)
    assert o_false.constant is False
    assert a._ops == {o_false.creator}
    assert b._ops == {o_false.creator}




import functools
import mygrad as mg

from hypothesis import given, note
import hypothesis.strategies as st
import hypothesis.extra.numpy as hnp

from numpy.testing import assert_allclose


def multi_matmul_slow(arrays): return functools.reduce(mg.matmul, arrays)


@given(num_arrays=st.integers(2, 5),
       left_1d=st.booleans(),
       right_1d=st.booleans(),
       output_is_constant=st.booleans(),
       data=st.data())
def test_multi_matmul(num_arrays, left_1d, right_1d, output_is_constant, data):
    """
    Ensures that ``multi_matmul`` behaves identically to:

        functools.reduce(mg.matmul, arrays)

    Includes edge cases in which the 1st and last tensors in the sequence are 1D
    """
    shape_endpoints = data.draw(st.tuples(*[st.integers(1, 10) for i in range(num_arrays + 1)]), label="endpoints")
    shapes = [shape_endpoints[i:i+2] for i in range(num_arrays)]

    if left_1d:
        shapes[0] = shapes[0][:0:-1]

    if right_1d:
        shapes[-1] = shapes[-1][:1]

    constants = data.draw(st.tuples(*[st.booleans() for i in range(num_arrays)]), label="constants")
    output_is_constant = output_is_constant or all(constants)

    arrs = [data.draw(hnp.arrays(dtype=float, shape=shapes[i], elements=st.floats(0, 1e6),),
                      label="arr-{}".format(i))
            for i in range(num_arrays)]
    note("tensor shapes: {}".format([i.shape for i in arrs]))

    arrs1 = [mg.Tensor(x, constant=const) for x, const in zip(arrs, constants)]
    arrs2 = [x.__copy__() for x in arrs1]

    actual = mg.multi_matmul(arrs1, constant=output_is_constant)
    desired = multi_matmul_slow(arrs2)
    assert_allclose(actual.data, desired.data, atol=1e-6, rtol=1e-6,
                    err_msg="`multi_matmul` does not produce the same result as "
                            "`functools.reduce(mg.matmul, arrays)`")

    assert actual.constant is output_is_constant, "`multi_matmul` does not carry constant info properly"

    if output_is_constant:
        return

    grad = data.draw(hnp.arrays(shape=desired.shape, dtype=float, elements=st.floats(0, 1e6)))

    (desired * grad).sum().backward()
    (actual * grad).sum().backward()

    for n, (const, arr1, arr2) in enumerate(zip(constants, arrs1, arrs2)):
        assert const is arr1.constant is arr2.constant, "tensor-{}-constant was not set properly".format(n)

        if const:
            assert arr2.grad is None, "tensor-{} is a constant, but its gradient is not `None`".format(n)
        else:
            assert_allclose(arr1.grad, arr2.grad, atol=1e-6, rtol=1e-6, err_msg="The gradients for tensor-{} for not match".format(n))

    actual.null_gradients()
    for n, arr1 in enumerate(arrs1):
        assert arr1.grad is None, "tensor-{} did not get its gradient nulled".format(n)




from tests.wrappers.uber import fwdprop_test_factory, backprop_test_factory
from mygrad import exp, expm1, exp2
import numpy as np


@fwdprop_test_factory(mygrad_func=exp, true_func=np.exp, num_arrays=1)
def test_exp_fwd():
    pass


@backprop_test_factory(mygrad_func=exp, true_func=np.exp, num_arrays=1)
def test_exp_backward():
    pass


@fwdprop_test_factory(mygrad_func=expm1, true_func=np.expm1, num_arrays=1)
def test_expm1_fwd():
    pass


@backprop_test_factory(mygrad_func=expm1, true_func=lambda x: np.exp(x) - 1, num_arrays=1)
def test_expm1_backward():
    pass


@fwdprop_test_factory(mygrad_func=exp2, true_func=np.exp2, num_arrays=1)
def test_exp2_fwd():
    pass


@backprop_test_factory(mygrad_func=exp2, true_func=np.exp2, num_arrays=1, as_decimal=False)
def test_exp2_backward():
    pass



import hypothesis.strategies as st
from hypothesis import given
from pytest import raises

from mygrad.tensor_base import Tensor
from mygrad.operation_base import Operation, BroadcastableOp

import numpy as np


class ScalarOnlyOp(BroadcastableOp):
    def __init__(self):
        self.scalar_only = True

    def __call__(self, a, b):
        self.variables = (a, b)
        return np.array([0])


class NotScalarOnlyOp(Operation):
    def __init__(self):
        self.scalar_only = False

    def __call__(self, a, b):
        self.variables = (a, b)
        return np.array([0])


@given(a_const=st.booleans(), a_scalar_only=st.booleans(),
       b_const=st.booleans(), b_scalar_only=st.booleans())
def test_scalar_only_op(a_const, a_scalar_only, b_const, b_scalar_only):
    """ op produces scalar_only result unless result is scalar. """
    a = Tensor(0, constant=a_const, _scalar_only=a_scalar_only)
    b = Tensor(0, constant=b_const, _scalar_only=b_scalar_only)

    out = Tensor._op(ScalarOnlyOp, a, b)
    scalar_only = True and not out.constant

    assert scalar_only is out.scalar_only

    # check out.backward()
    if scalar_only:
        with raises(Exception):
            out.backward()
    else:
        out.backward()  # a, b, out are const (nothing computed)


@given(a_const=st.booleans(), a_scalar_only=st.booleans(),
       b_const=st.booleans(), b_scalar_only=st.booleans())
def test_standard_op(a_const, a_scalar_only, b_const, b_scalar_only):
    """ op produces standard result unless an `a` or `b` is a scalar_only variable. """
    a = Tensor(0, constant=a_const, _scalar_only=a_scalar_only)
    b = Tensor(0, constant=b_const, _scalar_only=b_scalar_only)

    scalar_only = (a.scalar_only and not a.constant) or (b.scalar_only and not b.constant)
    out = Tensor._op(NotScalarOnlyOp, a, b)

    assert scalar_only is (out.scalar_only and not out.constant)

    # check out.backward()
    if scalar_only:
        with raises(Exception):
            out.backward()
    else:
        if a.constant and b.constant:
            out.backward()  # a, b, out are const (nothing computed)
        else:
            with raises(NotImplementedError):
                out.backward()


def test_practical_scalar_only():
    a = np.array([1, 2, 3])
    b = Tensor(3)
    out = a + b
    with raises(Exception):
        out.backward()

    out = a * b
    with raises(Exception):
        out.backward()

    out = a - b
    with raises(Exception):
        out.backward()

    out = a / b
    with raises(Exception):
        out.backward()

    out = a ** b
    with raises(Exception):
        out.backward()

    a = Tensor([1, 2, 3], constant=True)
    b = Tensor(3, constant=True)
    out = a + b
    out.backward()

    a = Tensor([1, 2, 3])
    b = 3
    out = a + b
    out.backward()

    out = a * b
    out.backward()

    out = a - b
    out.backward()

    out = a / b
    out.backward()

    out = a ** b
    out.backward()



from ..utils.numerical_gradient import numerical_gradient_full
from ..custom_strategies import broadcastable_shape

from mygrad import Tensor

from hypothesis import given
import hypothesis.strategies as st
import hypothesis.extra.numpy as hnp
from hypothesis import settings

from copy import copy

import numpy as np
from numpy.testing import assert_allclose
from itertools import chain
from mygrad.linalg.funcs import einsum


def bool_strat():
    """ einsum's optimize=True option has bugs prior to version 1.14.5
        (caught by these very unit tests!), thus we only test `optimize=True`
        for more recent versions."""
    return st.booleans() if np.__version__ <= "1.14.5" else st.just(False)


def compare_einsum(*operands, optimize=False):
    mygrad_out = einsum(*operands)
    assert isinstance(mygrad_out, Tensor)
    operands = tuple(i.data if isinstance(i, Tensor) else i for i in operands)
    assert_allclose(np.einsum(*operands), einsum(*operands, optimize=optimize).data)


def compare_backprop(*operands, atol=1e-5, rtol=1e-5, optimize=False):
    """ Compare back-propagation through mygrad-einsum, and compare
        against numerical derivative"""
    if isinstance(operands[0], str):
        # operands form: "ijk, ijk", x, y
        script = operands[0]
        vars = operands[1:]
        vars = tuple(np.asarray(i).astype(float) for i in vars)
        tensors = tuple(Tensor(i) for i in vars)

        def f(*args): return np.einsum(script, *args)
        out = einsum(script, *tensors, optimize=optimize)
    else:
        # operands form: op0, sublist0, op1, sublist1, ..., [sublistout]
        end = -1 if len(operands) % 2 else None  # -1 if sublistout is included
        vars = tuple(np.asarray(i).astype(float) for i in operands[:end:2])
        tensors = tuple(Tensor(i) for i in vars)

        def f(*args):
            x = tuple(chain.from_iterable(zip(args, operands[1::2])))
            if end is not None:
                x += (operands[-1],)
            return np.einsum(*x)
        x = tuple(chain.from_iterable(zip(tensors, operands[1::2])))
        if end is not None:
            x += (operands[-1],)
        out = einsum(*x, optimize=optimize)

    grad = np.random.rand(*out.shape)
    #    grad = np.ones(out.shape)
    out.backward(grad)

    numerical_derivs = numerical_gradient_full(f, *vars, back_grad=grad,
                                               as_decimal=False)

    for n, (dnum, tensor) in enumerate(zip(numerical_derivs, tensors)):
        assert dnum.shape == tensor.grad.shape
        assert_allclose(dnum, tensor.grad, atol=atol, rtol=rtol,
                        err_msg="The numerical and mygrad derivatives disagree for "
                                "variable index {}".format(n))


def test_unique_from_end():
    from mygrad.linalg.ops import _unique_from_end
    assert _unique_from_end("") == ""
    assert _unique_from_end("a") == "a"
    assert _unique_from_end("aaaa") == "a"
    assert _unique_from_end("aba") == "ba"
    assert _unique_from_end("abccbac") == "bac"


def test_merge_mappings():
    from mygrad.linalg.ops import _merge_max_mappings
    a = dict(a=0, b=100, c=3)
    b = dict(a=10, b=10)
    c = dict(c=50)
    d = dict(d=70)
    e = dict()
    assert _merge_max_mappings(a, b, c, d, e) == dict(a=10, b=100, c=50, d=70)


@given(optimize=bool_strat())
def test_einsum_static_fwd(optimize):
    """ Check all einsum examples from numpy doc"""
    a = Tensor(np.arange(25).reshape(5, 5))
    b = Tensor(np.arange(5))
    c = Tensor(np.arange(6).reshape(2, 3))

    compare_einsum('ii', a, optimize=optimize)
    compare_einsum(a, [0, 0], optimize=optimize)

    compare_einsum('ii->i', a, optimize=optimize)
    compare_einsum(a, [0, 0], [0], optimize=optimize)

    compare_einsum('ij,j', a, b, optimize=optimize)
    compare_einsum(a, [0, 1], b, [1], optimize=optimize)

    compare_einsum('...j,j', a, b, optimize=optimize)
    compare_einsum(a, [Ellipsis, 0], b, [Ellipsis, 0], optimize=optimize)

    compare_einsum('ji', c, optimize=optimize)
    compare_einsum(c, [1,0], optimize=optimize)

    compare_einsum('..., ...', 3, c, optimize=optimize)
    compare_einsum(3, [Ellipsis], c, [Ellipsis], optimize=optimize)

    compare_einsum('i,i', b, b, optimize=optimize)
    compare_einsum(b, [0], b, [0], optimize=optimize)

    compare_einsum('i,j', np.arange(2) + 1, b, optimize=optimize)
    compare_einsum('i...->...', a, optimize=optimize)

    a = np.arange(60.).reshape(3, 4, 5)
    b = np.arange(24.).reshape(4, 3, 2)
    compare_einsum('ijk,jil->kl', a, b, optimize=optimize)
    compare_einsum(a, [0, 1, 2], b, [1, 0, 3], [2, 3], optimize=optimize)

    a = np.arange(6).reshape((3, 2))
    b = np.arange(12).reshape((4, 3))
    compare_einsum('ki,jk->ij', a, b, optimize=optimize)
    compare_einsum(a, [0, 1], b, [2, 0], [1, 2], optimize=optimize)

    compare_einsum('ki,...k->i...', a, b, optimize=optimize)
    compare_einsum(a, [0, 1], b, [Ellipsis, 0], [1, Ellipsis], optimize=optimize)

    compare_einsum('k...,jk', a, b, optimize=optimize)
    compare_einsum(a, [0, Ellipsis], b, [2, 0], optimize=optimize)


@given(optimize=bool_strat())
def test_einsum_static_bkwd(optimize):
    """ Check all einsum examples from numpy doc"""
    a = np.arange(25).reshape(5, 5)
    b = np.arange(5)
    c = np.arange(6).reshape(2, 3)

    compare_backprop('ii', a, optimize=optimize)
    compare_backprop(a, [0, 0], optimize=optimize)

    compare_backprop('ii->i', a, optimize=optimize)
    compare_backprop(a, [0, 0], [0], optimize=optimize)

    compare_backprop('ij->', a, optimize=optimize)
    compare_backprop(a, [0, 1], optimize=optimize)

    compare_backprop('ij,j', a, b, optimize=optimize)
    compare_backprop(a, [0, 1], b, [1], optimize=optimize)

    compare_backprop('...j,j', a, b, optimize=optimize)
    compare_backprop(a, [Ellipsis, 0], b, [Ellipsis, 0], optimize=optimize)
    
    compare_backprop('ji', c, optimize=optimize)
    compare_backprop(c, [1, 0], optimize=optimize)
    
    compare_backprop('..., ...', 3, c, optimize=optimize)
    compare_backprop(3, [Ellipsis], c, [Ellipsis], optimize=optimize)
    
    compare_backprop('i,i', b, b, optimize=optimize)
    compare_backprop(b, [0], b, [0], optimize=optimize)
    
    compare_backprop('i,j', np.arange(2) + 1, b, optimize=optimize)

    a = np.arange(60.).reshape(3, 4, 5)
    b = np.arange(24.).reshape(4, 3, 2)
    compare_backprop('ijk,jil->kl', a, b, atol=1e-3, rtol=1e-3, optimize=optimize)
    compare_backprop(a, [0, 1, 2], b, [1, 0, 3], [2, 3], atol=1e-3, rtol=1e-3, optimize=optimize)
    
    a = np.arange(6).reshape((3, 2))
    b = np.arange(4).reshape((4, 1))
    compare_backprop('ki,jk->ij', a, b, optimize=optimize)
    compare_backprop(a, [0, 1], b, [2, 0], [1, 2], optimize=optimize)
    
    compare_backprop('ki,...k->i...', a, b, optimize=optimize)
    compare_backprop(a, [0, 1], b, [Ellipsis, 0], [1, Ellipsis], optimize=optimize)
    
    compare_backprop('k...,jk', a, b, optimize=optimize)
    compare_backprop(a, [0, Ellipsis], b, [2, 0], optimize=optimize)


@settings(deadline=1000)
@given(optimize=bool_strat())
def test_traces_bkwd(optimize):
    a = np.random.rand(5, 2, 2, 5)
    b = np.random.rand(3, 2, 1)
    c = np.random.rand(1, 1)
    d = np.random.rand(5, 5, 5)
    compare_backprop('ijji -> i', a, optimize=optimize)
    compare_backprop(a, [0, 1, 1, 0], [0], optimize=optimize)

    compare_backprop("iii -> i", d, optimize=optimize)
    compare_backprop('ijji -> j', a, optimize=optimize)
    compare_backprop('ijji -> ij', a, optimize=optimize)
    compare_backprop('ijji -> ji', a, optimize=optimize)
    compare_backprop('ijji -> ', a, optimize=optimize)
    compare_backprop('ijji,kji -> ', a, b, optimize=optimize)
    compare_backprop('ijji,kji -> kj', a, b, optimize=optimize)
    compare_backprop('ijji,kji,jj-> kj', a, b, c, optimize=optimize)
    compare_backprop('ijji,kji,jj-> ijk', a, b, c, optimize=optimize)
    compare_backprop('ijji,kji,jj-> jk', a, b, c, optimize=optimize)


def test_redundant_args():
    """
    Test behavior for when einsum receives redundant inputs. An optimization
    was added such that einsum will only compute the gradient for such an entry
    once and scale it accordingly.
    """
    a = Tensor(np.arange(4).reshape(2, 2))
    a_copy = copy(a)

    # check standard summation
    o = einsum("ij,ij", a, a)
    assert len(o.creator.cache) == 1
    o.sum().backward()

    o = einsum("ij,ij", a_copy, a_copy*1)
    assert len(o.creator.cache) == 2
    o.sum().backward()
    assert_allclose(a.grad, a_copy.grad)

    a = Tensor(np.arange(4).reshape(2, 2))
    a_copy = copy(a)

    # check standard summation using alt signature
    o = einsum(a, [0, 1], a, [0, 1])
    assert len(o.creator.cache) == 1
    o.sum().backward()

    o = einsum(a_copy, [0, 1], a_copy*1, [0, 1])
    assert len(o.creator.cache) == 2
    o.sum().backward()
    assert_allclose(a.grad, a_copy.grad)

    a = Tensor(np.arange(4).reshape(2, 2))
    a_copy = copy(a)

    # check matmul (no redundant indices)
    o = einsum("ij,jk", a, a)
    assert len(o.creator.cache) == 2
    o.sum().backward()

    o = a_copy @ a_copy
    o.sum().backward()
    assert_allclose(a.grad, a_copy.grad)

    a = Tensor(np.arange(4).reshape(2, 2))
    a_copy = copy(a)

    # check traces
    o = einsum("ii,ii", a, a)
    assert len(o.creator.cache) == 1
    o.sum().backward()

    o = einsum("ii,ii", a_copy, a_copy*1)
    assert len(o.creator.cache) == 2
    o.sum().backward()
    assert_allclose(a.grad, a_copy.grad)

    a = Tensor(np.arange(4).reshape(2, 2))
    a_copy = copy(a)

    b = Tensor(-1*np.arange(2).reshape(2, 1))
    b_copy = copy(b)

    # check broadcasting and multiply-redundant input tensors
    # with distinct einsum labels
    o = einsum("ii,ii,i...,i...,...i,...i", a, a, b, b, a, a)
    assert len(o.creator.cache) == 3
    o.sum().backward()

    o = einsum("ii,ii,i...,i...,...i,...i", a_copy, a_copy*1, b_copy, b_copy*1, a_copy, 1*a_copy)
    assert len(o.creator.cache) == 6
    o.sum().backward()
    assert_allclose(a.grad, a_copy.grad)
    assert_allclose(b.grad, b_copy.grad)


@given(num=st.integers(1, 10),
       optimize=bool_strat(),
       data=st.data())
def test_einsum_bkwd1(num, optimize, data):
    x = Tensor(np.random.rand(num))
    y_shape = data.draw(broadcastable_shape(x.shape, min_dim=1, max_dim=1))
    y = Tensor(np.random.rand(*y_shape))

    grad = data.draw(st.floats(-100, 100))
    o = einsum("i, i", x, y, optimize=optimize)
    o.backward(grad)

    def f(x, y): return np.einsum("i, i", x, y)

    dx, dy = numerical_gradient_full(f, x.data, y.data, back_grad=grad, as_decimal=False)

    assert_allclose(x.grad, dx, atol=1e-5, rtol=1e-5)
    assert_allclose(y.grad, dy, atol=1e-5, rtol=1e-5)

    o.null_gradients()
    assert x.grad is None
    assert y.grad is None

    # test broadcasting in reverse direction
    o = einsum("i, i", y, x, optimize=optimize)
    o.backward(grad)

    dy, dx = numerical_gradient_full(f, y.data, x.data, back_grad=grad, as_decimal=False)

    assert_allclose(x.grad, dx, atol=1e-5, rtol=1e-5)
    assert_allclose(y.grad, dy, atol=1e-5, rtol=1e-5)

    o.null_gradients()


@given(num=st.integers(1, 10),
       optimize=bool_strat(),
       data=st.data())
def test_einsum_bkwd2(num, optimize, data):
    y = Tensor(np.random.rand(num))

    # flip so that leading dim of x is broadcastable with y
    x_shape = data.draw(broadcastable_shape(y.shape, min_dim=2, max_dim=2))[::-1]
    x = Tensor(np.random.rand(*x_shape))
    grad = np.random.rand(x.shape[-1])

    o = einsum("ia, i -> a", x, y, optimize=optimize)
    o.backward(grad)

    def f(x, y): return np.einsum("ia, i -> a", x, y)

    dx, dy = numerical_gradient_full(f, x.data, y.data, back_grad=grad, as_decimal=False)

    assert_allclose(x.grad, dx, atol=1e-6)
    assert_allclose(y.grad, dy, atol=1e-6)


@given(shape=hnp.array_shapes(min_dims=2, max_dims=2),
       optimize=bool_strat(),
       data=st.data())
def test_einsum_bkwd3(shape, optimize, data):
    script = "ia, ia, i -> a"
    x = Tensor(np.random.rand(*shape))

    y_shape = data.draw(broadcastable_shape(shape, min_dim=2, max_dim=2))
    y = Tensor(np.random.rand(*y_shape))

    z_shape = data.draw(broadcastable_shape(x.shape[:1], min_dim=1, max_dim=1))
    z = Tensor(np.random.rand(*z_shape ))

    grad = np.random.rand(x.shape[1])

    o = einsum(script, x, y, z, optimize=optimize)
    o.backward(grad)

    def f(x, y, z): return np.einsum(script, x, y, z)

    dx, dy, dz = numerical_gradient_full(f, x.data, y.data, z.data, back_grad=grad,
                                         as_decimal=False)

    assert_allclose(x.grad, dx, atol=1e-6)
    assert_allclose(y.grad, dy, atol=1e-6)
    assert_allclose(z.grad, dz, atol=1e-6)


@given(shape=hnp.array_shapes(min_dims=2, max_dims=2),
       optimize=bool_strat(),
       data=st.data())
def test_einsum_bkwd4(shape, optimize, data):
    script = "ia, i -> "

    x = Tensor(np.random.rand(*shape))

    y_shape = data.draw(broadcastable_shape(x.shape[:1], min_dim=1, max_dim=1))
    y = Tensor(np.random.rand(*y_shape))

    grad = np.random.rand(1).item()

    o = einsum(script, x, y, optimize=optimize)
    o.backward(grad)

    def f(x, y): return np.einsum(script, x, y)

    dx, dy = numerical_gradient_full(f, x.data, y.data, back_grad=grad,
                                     as_decimal=False)

    assert_allclose(x.grad, dx, atol=1e-6)
    assert_allclose(y.grad, dy, atol=1e-6)


@settings(deadline=2000)
@given(optimize=bool_strat())
def test_einsum_bkwd5(optimize):
    x = Tensor(np.random.rand(5, 3, 4, 6))
    y = Tensor(np.random.rand(1, 5, 6, 2))
    grad = np.random.rand(1, 3, 4, 2)

    def f(x, y): return np.einsum("iBCj, aijd -> aBCd", x, y)

    o = einsum("iBCj, aijd -> aBCd", x, y, optimize=optimize)
    o.backward(grad)

    dx, dy = numerical_gradient_full(f, x.data, y.data, back_grad=grad,
                                     as_decimal=False)

    assert_allclose(x.grad, dx, atol=1e-6)
    assert_allclose(y.grad, dy, atol=1e-6)


@settings(deadline=2000)
@given(shape=hnp.array_shapes(min_dims=3, max_dims=3),
       optimize=bool_strat())
def test_einsum_bkwd6(shape, optimize):
    sig = "ijk, -> j"
    x = Tensor(np.random.rand(*shape))
    y = Tensor(np.random.rand(1).item())
    grad = np.random.rand(x.shape[1])

    o = einsum(sig, x, y, optimize=optimize)
    o.backward(grad)

    def f(x, y): return np.einsum(sig, x, y)

    dx, dy = numerical_gradient_full(f, x.data, y.data, back_grad=grad,
                                     as_decimal=False)

    assert_allclose(x.grad, dx, atol=1e-6)
    assert_allclose(y.grad, dy, atol=1e-6)



from mygrad import Tensor
from mygrad.operation_base import Operation
import mygrad as mg

from hypothesis import given
import hypothesis.strategies as st
import hypothesis.extra.numpy as hnp

from pytest import raises
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal, assert_equal


def test_to_scalar():
    nd_tensor = Tensor([1, 2])
    with raises(TypeError):
        float(nd_tensor)

    with raises(TypeError):
        int(nd_tensor)

    with raises(ValueError):
        nd_tensor.item()

    for size1_tensor in (Tensor(1), Tensor([[1]])):
        assert float(size1_tensor) == 1.
        assert int(size1_tensor) == 1
        assert size1_tensor.item() == 1.


def test_repr():
    assert repr(Tensor(1)) == 'Tensor(1)'
    assert repr(Tensor([1])) == 'Tensor([1])'
    assert repr(Tensor([1, 2])) == 'Tensor([1, 2])'
    tmp_rep = 'Tensor([[0, 1, 2],\n        [3, 4, 5],\n        [6, 7, 8]])'
    assert repr(mg.arange(9).reshape((3, 3))) == tmp_rep


def test_contains():
    t = Tensor([[0, 1, 2], [3, 4, 5]])
    assert 0 in t and 0 in t.data
    assert [0, 1, 2] in t and [0, 1, 2] in t.data
    assert -1 not in t and -1 not in t.data


@given(a=hnp.arrays(shape=hnp.array_shapes(max_side=3, max_dims=5),
                    dtype=float,
                    elements=st.floats(-100, 100)),
       constant=st.booleans(),
       scalar=st.booleans(),
       creator=st.booleans())
def test_properties(a, constant, scalar, creator):
    array = np.asarray(a)
    if creator:
        ref = Operation()
        tensor = Tensor(a, constant=constant, _creator=ref, _scalar_only=scalar)
    else:
        tensor = Tensor(a, constant=constant, _scalar_only=scalar)

    assert tensor.ndim == array.ndim
    assert tensor.shape == array.shape
    assert tensor.size == array.size
    assert len(tensor) == len(array)
    assert tensor.dtype == array.dtype
    assert_equal(actual=tensor.data, desired=a)
    assert (not creator) or tensor.creator is ref


def test_init_data():
    for data in [0, [], (0, 0), ((0, 0), (0, 0)), np.random.rand(3, 4, 2)]:
        assert_equal(actual=Tensor(data).data, desired=np.asarray(data),
                     err_msg="Initialization with non-tensor failed")
        assert_equal(actual=Tensor(Tensor(data)).data, desired=np.asarray(data),
                     err_msg="Initialization with tensor failed")


@given(x=hnp.arrays(dtype=float, shape=hnp.array_shapes(min_dims=1, max_dims=4)))
def test_init_data_rand(x):
    assert_equal(actual=Tensor(x).data, desired=x)


@given(x=hnp.arrays(dtype=float, shape=hnp.array_shapes()))
def test_items(x):
    """ verify that tensor.item() mirrors array.item()"""
    tensor = Tensor(x)
    try:
        value = x.item()
        assert_allclose(value, tensor.item())
    except ValueError:
        with raises(ValueError):
            tensor.item()


op = Operation()
dtype_strat = st.sampled_from((None, int, float,
                               np.int8, np.int16, np.int32, np.int64,
                               np.float16, np.float32, np.float64))
dtype_strat_numpy = st.sampled_from((np.int8, np.int16, np.int32, np.int64,
                                     np.float16, np.float32, np.float64))


@given(data=st.data(),
       creator=st.sampled_from((None, op)),
       constant=st.booleans(),
       scalar_only=st.booleans(),
       dtype=dtype_strat,
       numpy_dtype=dtype_strat_numpy)
def test_init_params(data, creator, constant, scalar_only, dtype, numpy_dtype):
    elements = st.floats if np.issubdtype(numpy_dtype, np.floating) else st.integers
    a = data.draw(hnp.arrays(shape=hnp.array_shapes(max_side=3, max_dims=5),
                             dtype=numpy_dtype,
                             elements=elements(-100, 100)),
                  label="a")
    if dtype is not None:
        a = a.astype(dtype)

    tensor = Tensor(a, _creator=creator, constant=constant, _scalar_only=scalar_only, dtype=dtype)

    assert tensor.creator is creator
    assert tensor.constant is constant
    assert tensor.scalar_only is scalar_only
    assert tensor.dtype is a.dtype
    assert_equal(tensor.data, a)
    assert tensor.grad is None


@given(constant_x=st.booleans(), constant_y=st.booleans())
def test_special_methods(constant_x: bool, constant_y: bool):
    from mygrad.math.arithmetic.ops import Add, Subtract, Multiply, Divide, Power
    from mygrad.math.arithmetic.ops import Negative
    from mygrad.linalg.ops import MatMul

    x = Tensor([2., 8., 5.], constant=constant_x)
    y = Tensor([1., 3., 2.], constant=constant_y)

    constant = constant_x and constant_y

    for op_name, op in zip(("__add__", "__sub__", "__mul__", "__truediv__", "__pow__", "__matmul__"),
                           (Add, Subtract, Multiply, Divide, Power, MatMul)):
        assert hasattr(Tensor, op_name), "`Tensor` is missing the attribute {}".format(op_name)
        tensor_out = getattr(Tensor, op_name)(x, y)
        numpy_out = getattr(np.ndarray, op_name)(x.data, y.data)
        assert isinstance(tensor_out, Tensor), "`Tensor.{}` did not return a Tensor-instance".format(op_name)

        assert tensor_out.constant is constant, "`Tensor.{}` did not propagate constant properly".format(op_name)

        assert_equal(tensor_out.data, numpy_out, err_msg="`Tensor.{}` did not produce the correct "
                                                         "numerical result.".format(op_name))

        assert isinstance(tensor_out.creator, op), "`Tensor.{}` does not have the expected " \
                                                   "creator-instance".format(op_name)

        assert tensor_out.creator.variables[0] is x, "`Tensor.{}`'s creator did not record the " \
                                                     "correct variable x".format(op_name)

        assert tensor_out.creator.variables[1] is y, "`Tensor.{}`'s creator did not record the " \
                                                     "correct variable y".format(op_name)

    for op_name, op in zip(("__radd__", "__rsub__", "__rmul__", "__rtruediv__", "__rpow__", "__rmatmul__"),
                           (Add, Subtract, Multiply, Divide, Power, MatMul)):
        assert hasattr(Tensor, op_name), "`Tensor` is missing the attribute {}".format(op_name)
        tensor_out = getattr(Tensor, op_name)(x, y)
        numpy_out = getattr(np.ndarray, op_name)(x.data, y.data)
        assert isinstance(tensor_out, Tensor), "`Tensor.{}` did not return a Tensor-instance".format(op_name)
        assert tensor_out.constant is constant, "`Tensor.{}` did not propagate constant properly".format(op_name)
        assert_equal(tensor_out.data, numpy_out, err_msg="`Tensor.{}` did not produce the correct "
                                                         "numerical result.".format(op_name))
        assert isinstance(tensor_out.creator, op), "`Tensor.{}` does not have the expected " \
                                                   "creator-instance".format(op_name)

        assert tensor_out.creator.variables[0] is y, "`Tensor.{}`'s creator did not record the " \
                                                     "correct variable x".format(op_name)

        assert tensor_out.creator.variables[1] is x, "`Tensor.{}`'s creator did not record the " \
                                                     "correct variable y".format(op_name)

    op_name = "__neg__"
    assert hasattr(Tensor, op_name), "`Tensor` is missing the attribute {}".format(op_name)
    tensor_out = getattr(Tensor, "__neg__")(x)
    numpy_out = getattr(np.ndarray, "__neg__")(x.data)
    assert isinstance(tensor_out, Tensor), "`Tensor.{}` did not return a Tensor-instance".format(op_name)

    assert_equal(tensor_out.data, numpy_out, err_msg="`Tensor.{}` did not produce the correct "
                                                     "numerical result.".format(op_name))

    assert isinstance(tensor_out.creator, Negative), "`Tensor.{}` does not have the expected " \
                                                     "creator-instance".format(op_name)

    assert tensor_out.creator.variables[0] is x, "`Tensor.{}`'s creator did not record the " \
                                                 "correct variable x".format(op_name)


def test_comparison_ops():
    x = Tensor([1, 3, 5])
    y = Tensor([1, 4, 2])
    for op in ("__lt__", "__le__", "__gt__", "__ge__", "__eq__", "__ne__"):
        assert hasattr(Tensor, op), "`Tensor` is missing the attribute {}".format(op)
        tensor_out = getattr(Tensor, op)(x, y)
        array_out = getattr(np.ndarray, op)(x.data, y.data)
        assert_equal(actual=tensor_out, desired=array_out, err_msg="The comparison operation `Tensor.{}` "
                                                                   "does not match the expected behavior "
                                                                   "by comparison to numpy.ndarray".format(op))


@given(constant=st.booleans())
def test_math_methods(constant: bool):
    x = Tensor([[1., 2., 3.],
                [4., 5., 6.]], constant=constant)
    for attr in ("sum", "prod", "cumprod", "cumsum",
                 "mean", "std", "var",
                 "max", "min",
                 "transpose", "squeeze", "ravel"):
        assert hasattr(x, attr), "`Tensor` is missing the attribute: {attr}".format(attr=attr)
        method_out = getattr(x, attr).__call__()
        function_out = getattr(mg, attr).__call__(x)
        assert_equal(method_out.data, function_out.data)
        assert method_out.constant is constant, "`Tensor.{}` did not propagate constant properly".format(attr)
        assert type(method_out.creator) is type(function_out.creator)

    method_out = x.moveaxis(0, -1)
    function_out = mg.moveaxis(x, 0, -1)
    assert_equal(method_out.data, function_out.data)
    assert method_out.constant is constant, "`Tensor.moveaxis` did not propagate constant properly".format(attr)
    assert type(method_out.creator) is type(function_out.creator)

    method_out = x.swapaxes(0, -1)
    function_out = mg.swapaxes(x, 0, -1)
    assert_equal(method_out.data, function_out.data)
    assert method_out.constant is constant, "`Tensor.swapaxes` did not propagate constant properly".format(attr)
    assert type(method_out.creator) is type(function_out.creator)


@given(x=st.floats(min_value=-1E6, max_value=1E6),
       y=st.floats(min_value=-1E6, max_value=1E6),
       z=st.floats(min_value=-1E6, max_value=1E6),
       clear_graph=st.booleans())
def test_null_gradients(x, y, z, clear_graph):
    x = Tensor(x)
    y = Tensor(y)
    z = Tensor(z)

    f = x*y + z
    g = x + z*f*f

    # check side effects
    unused = 2*g - f
    w = 1*f

    g.backward()
    assert x.grad is not None
    assert y.grad is not None
    assert z.grad is not None
    assert f.grad is not None
    assert g.grad is not None
    assert len(x._ops) > 0
    assert len(y._ops) > 0
    assert len(z._ops) > 0
    assert len(f._ops) > 0
    assert len(g._ops) > 0
    assert w.grad is None

    g.null_gradients(clear_graph=clear_graph)
    assert x.grad is None
    assert y.grad is None
    assert z.grad is None
    assert f.grad is None
    assert g.grad is None

    if clear_graph:
        assert len(x._ops) == 0
        assert len(y._ops) == 0
        assert len(z._ops) == 0
        assert len(f._ops) == 0
        assert len(g._ops) > 0
        assert x.creator is None
        assert y.creator is None
        assert z.creator is None
        assert f.creator is None
        assert g.creator is None
    else:
        assert len(x._ops) > 0
        assert len(y._ops) > 0
        assert len(z._ops) > 0
        assert len(f._ops) > 0
        assert len(g._ops) > 0
        assert x.creator is None
        assert y.creator is None
        assert z.creator is None
        assert f.creator is not None
        assert g.creator is not None


@given(x=st.floats(min_value=-1E-6, max_value=1E6),
       y=st.floats(min_value=-1E-6, max_value=1E6),
       z=st.floats(min_value=-1E-6, max_value=1E6))
def test_clear_graph(x, y, z):
    x_orig = x
    y_orig = y
    z_orig = z

    x = Tensor(x)
    y = Tensor(y)
    z = Tensor(z)

    f = x*y + z
    g = x + z*f*f

    # check side effects
    unused = 2*g - f
    w = 1*f

    g.backward()
    assert_allclose(f.grad, 2 * z.data * f.data)
    assert_allclose(x.grad, 1 + 2 * z.data * f.data * y.data)
    assert_allclose(y.grad, 2 * z.data * f.data * x.data)
    assert_allclose(z.grad, f.data**2 + z.data * 2 * f.data)
    assert w.grad is None

    assert_array_equal(x.data, x_orig, err_msg="x was mutated during the operation")
    assert_array_equal(y.data, y_orig, err_msg="y was mutated during the operation")
    assert_array_equal(z.data, z_orig, err_msg="z was mutated during the operation")

    # null-gradients without clearing the graph, confirm that backprop still works
    g.null_gradients(clear_graph=False)
    g.backward()
    assert_allclose(f.grad, 2 * z.data * f.data)
    assert_allclose(x.grad, 1 + 2 * z.data * f.data * y.data)
    assert_allclose(y.grad, 2 * z.data * f.data * x.data)
    assert_allclose(z.grad, f.data**2 + z.data * 2 * f.data)
    assert w.grad is None

    assert_array_equal(x.data, x_orig, err_msg="x was mutated during the operation")
    assert_array_equal(y.data, y_orig, err_msg="y was mutated during the operation")
    assert_array_equal(z.data, z_orig, err_msg="z was mutated during the operation")

    g.null_gradients(clear_graph=False)
    w.backward()
    assert_allclose(x.grad, y.data)
    assert_allclose(y.grad, x.data)
    assert_allclose(z.grad, np.array(1.))

    w.clear_graph()
    assert_allclose(x.grad, y.data)
    assert_allclose(y.grad, x.data)
    assert_allclose(z.grad, np.array(1.))
    assert len(g._ops) > 0
    assert g.creator is not None
    assert len(x._ops) == 0
    assert len(y._ops) == 0
    assert len(z._ops) == 0
    assert len(f._ops) == 0
    assert x.creator is None
    assert y.creator is None
    assert z.creator is None
    assert f.creator is None

    with raises(Exception):
        g.backward()



import os
from hypothesis import settings, Verbosity

settings.register_profile("ci", deadline=1000)
settings.register_profile("intense", deadline=None, max_examples=1000)
settings.register_profile("dev", max_examples=10)
settings.register_profile("debug", max_examples=10, verbosity=Verbosity.verbose)
settings.load_profile(os.getenv(u'HYPOTHESIS_PROFILE', 'default'))



from mygrad.operation_base import Operation, BroadcastableOp
import numpy as np

__all__ = ["Sin",
           "Sinc",
           "Cos",
           "Tan",
           "Csc",
           "Sec",
           "Cot",
           "Arcsin",
           "Arccos",
           "Arctan",
           "Arccsc",
           "Arcsec",
           "Arccot",
           "Arctan2"]


class Sin(Operation):
    """ f(a) -> sin(a)"""
    def __call__(self, a):
        self.variables = (a,)
        return np.sin(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad * np.cos(a.data)


def _dsinc(x):
    x = x * np.pi
    return (x * np.cos(x) - np.sin(x)) / x ** 2


class Sinc(Operation):
    """ f(a) -> sin(pi*a)/(pi*a)"""
    def __call__(self, a):
        self.variables = (a,)
        return np.sinc(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        x = a.data
        return np.pi * grad * np.piecewise(x, [x == 0, x != 0], [np.zeros_like, _dsinc])


class Cos(Operation):
    """ f(a) -> cos(a)"""
    def __call__(self, a):
        self.variables = (a,)
        return np.cos(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad * -np.sin(a.data)


class Tan(Operation):
    """ f(a) -> tan(a)"""
    def __call__(self, a):
        self.variables = (a,)
        return np.tan(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad / np.cos(a.data)**2


class Csc(Operation):
    """ f(a) -> csc(a)"""
    def __call__(self, a):
        self.variables = (a,)
        return 1 / np.sin(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad * -np.cos(a.data) / np.sin(a.data)**2


class Sec(Operation):
    """ f(a) -> sec(a)"""
    def __call__(self, a):
        self.variables = (a,)
        return 1 / np.cos(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad * np.sin(a.data) / np.cos(a.data)**2


class Cot(Operation):
    """ f(a) -> cot(a)"""
    def __call__(self, a):
        self.variables = (a,)
        return 1 / np.tan(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return -grad / np.sin(a.data)**2

class Arcsin(Operation):
    """ f(a) -> arcsin(a)"""
    def __call__(self, a):
        self.variables = (a,)
        if np.any(1 < a) or np.any(a < -1):
            raise ValueError("Invalid arcsin-domain value")
        return np.arcsin(a.data)

    def backward_var(self, grad, index, **kwargs):
        # d arcsin / dx at x = -1, 1 returns 0, not NaN
        a = self.variables[index]
        return np.select([np.abs(a.data) != 1], [grad / np.sqrt(1 - a.data ** 2)])


class Arccos(Operation):
    """ f(a) -> arccos(a)"""
    def __call__(self, a):
        self.variables = (a,)
        if np.any(1 < a) or np.any(a < -1):
            raise ValueError("Invalid arccos-domain value")
        return np.arccos(a.data)

    def backward_var(self, grad, index, **kwargs):
        # d arccos / dx at x = -1, 1 returns 0, not NaN
        a = self.variables[index]
        return np.select([np.abs(a.data) != 1], [-grad / np.sqrt(1 - a.data ** 2)])


class Arctan(Operation):
    """ f(a) -> arctan(a)"""
    def __call__(self, a):
        self.variables = (a,)
        return np.arctan(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad / (1 + a.data ** 2)


class Arccsc(Operation):
    """ f(a) -> arccsc(a)"""
    def __call__(self, a):
        self.variables = (a,)
        if np.any(-1 < a) and np.any(a < 1):
            raise ValueError("Invalid arccsc-domain value")
        return np.arcsin(1 / a.data)

    def backward_var(self, grad, index, **kwargs):
        # d arccsc / dx at x = -1, 1 returns 0, not NaN
        a = self.variables[index]
        return np.select([np.abs(a.data) != 1], [-grad / (np.abs(a.data) * np.sqrt(a.data ** 2 - 1))])


class Arcsec(Operation):
    """ f(a) -> arcsec(a)"""
    def __call__(self, a):
        self.variables = (a,)
        if np.any(-1 < a) and np.any(a < 1):
            raise ValueError("Invalid arcsec-domain value")
        return np.arccos(1 / a.data)

    def backward_var(self, grad, index, **kwargs):
        # d arcsec / dx at x = -1, 1 returns 0, not NaN
        a = self.variables[index]
        return np.select([np.abs(a.data) != 1], [grad / (np.abs(a.data) * np.sqrt(a.data ** 2 - 1))])


class Arccot(Operation):
    """ f(a) -> arccot(a)"""
    def __call__(self, a):
        self.variables = (a,)
        return np.piecewise(a.data, [a.data == 0, a.data != 0], [np.pi / 2, lambda x: np.arctan(1 / x)])

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return -grad / (1 + a.data ** 2)

class Arctan2(BroadcastableOp):
    """ f(a, b) -> arctan(a/b)"""
    def __call__(self, a, b):
        self.variables = (a,b)
        return np.arctan2(a.data, b.data)

    def backward_var(self, grad, index, **kwargs):
        a, b = self.variables
        if index == 0:
            return grad * b.data / (a.data ** 2 + b.data ** 2)
        else:
            return -1.0 * grad * a.data / (a.data ** 2 + b.data ** 2)


import mygrad as mg
from mygrad.tensor_base import Tensor
from mygrad.nnet.losses import multiclass_hinge

import numpy as np
from numpy.testing import assert_allclose
import hypothesis.strategies as st
from hypothesis import given
import hypothesis.extra.numpy as hnp

from pytest import raises

@given(st.data())
def test_multiclass_hinge(data):
    """ Test the built-in implementation of multiclass hinge against the pure pygrad version"""
    s = data.draw(hnp.arrays(shape=hnp.array_shapes(max_side=10, min_dims=2, max_dims=2),
                             dtype=float,
                             elements=st.floats(-100, 100)))
    l = data.draw(hnp.arrays(shape=(s.shape[0],),
                             dtype=int,
                             elements=st.integers(min_value=0, max_value=s.shape[1] - 1)))
    hinge_scores = Tensor(s)
    hinge_loss = multiclass_hinge(hinge_scores, l, constant=False)
    hinge_loss.backward()

    pygrad_scores = Tensor(s)
    correct_labels = (range(len(l)), l)
    correct_class_scores = pygrad_scores[correct_labels]  # Nx1

    Lij = pygrad_scores - correct_class_scores[:, np.newaxis] + 1.  # NxC margins
    Lij[Lij <= 0] = 0
    Lij[correct_labels] = 0

    pygrad_loss = Lij.sum() / pygrad_scores.shape[0]
    pygrad_loss.backward()
    assert_allclose(hinge_loss.data, pygrad_loss.data)
    assert_allclose(pygrad_scores.grad, hinge_scores.grad)


@given(shape=st.sampled_from([(3, 1), (3, 4), tuple()]))
def test_bad_label_shape(shape):
    """
    Ensures that `multiclass_hinge` checks for shape-(N,) `y_true`
    """
    scores = mg.arange(12).reshape(3, 4)
    labels = mg.zeros(shape, dtype=int)
    with raises(ValueError):
        multiclass_hinge(scores, labels)


@given(type=st.sampled_from([bool, float, np.float32]))
def test_bad_label_type(type):
    """
    Ensures that `multiclass_hinge` checks integer-type `y_true`
    """
    scores = mg.arange(12).reshape(3, 4)
    labels = np.zeros((3,), dtype=type)
    with raises(TypeError):
        multiclass_hinge(scores, labels)



from .ops import Add, Subtract, Power, Square, Divide, Multiply, MultiplySequence, AddSequence
from .ops import Positive, Negative, Reciprocal
from mygrad.tensor_base import Tensor


__all__ = ["add",
           "subtract",
           "power",
           "divide",
           "square",
           "reciprocal",
           "multiply",
           "multiply_sequence",
           "add_sequence",
           "positive",
           "negative"]


def add(a, b, constant=False):
    """ ``f(a, b) -> a + b``

        Parameters
        ----------
        a : array_like

        b : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor
        
        Examples
        --------
        >>> import mygrad as mg
        >>> mg.add(1.0, 4.0)
        Tensor(5.0)
        >>> x1 = mg.arange(9.0).reshape((3, 3))
        >>> x2 = mg.arange(3.0)
        >>> mg.add(x1, x2)  # equivalent to `x1 + x2`
        Tensor([[  0.,   2.,   4.],
                [  3.,   5.,   7.],
                [  6.,   8.,  10.]])"""
    return Tensor._op(Add, a, b, constant=constant)


def subtract(a, b, constant=False):
    """ ``f(a, b) -> a - b``

        Parameters
        ----------
        a : array_like

        b : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Examples
        --------
        >>> import mygrad as mg
        >>> mg.subtract(1.0, 4.0, constant=True)  # resulting tensor will not back-propagate a gradient
        Tensor(5.0)
        >>> x1 = mg.arange(9.0).reshape((3, 3))
        >>> x2 = mg.arange(3.0)
        >>> mg.subtract(x2, x1)  # equivalent to `x2 - x1`
        Tensor([[  0.,   0.,   0.],
                [  3.,   3.,   3.],
                [  6.,   6.,  6.]])
    """
    return Tensor._op(Subtract, a, b, constant=constant)


def divide(a, b, constant=False):
    """ ``f(a, b) -> a / b``

        Parameters
        ----------
        a : array_like

        b : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Divide, a, b, constant=constant)

def square(a, constant=False):
    """ ``f(a) -> a ** 2``

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Square, a, constant=constant)


def reciprocal(a, constant=False):
    """ ``f(a) -> 1 / a``

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Reciprocal, a, constant=constant)


def power(a, b, constant=False):
    """ ``f(a, b) -> a ** b``

        Parameters
        ----------
        a : array_like

        b : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Power, a, b, constant=constant)


def multiply(a, b, constant=False):
    """ ``f(a, b) -> a * b``

        Parameters
        ----------
        a : array_like

        b : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Multiply, a, b, constant=constant)


def multiply_sequence(*variables, constant=False):
    """  ``f(a, b, ...) -> a + b + ...``

        Add a sequence of N tensors.

        Parameters
        ----------
        variables : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Notes
        -----
        It is more efficient to back-propagate through this
        function than it is through a computational graph
        with N-1 corresponding addition operations."""
    return Tensor._op(MultiplySequence, *variables, constant=constant)


def add_sequence(*variables, constant=False):
    """ ``f(a, b, ...) -> a * b * ...``

        Multiply a sequence of N tensors.

        Parameters
        ----------
        variables : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Notes
        -----
        It is more efficient to back-propagate through this
        function than it is through a computational graph
        with N-1 corresponding multiplication operations."""
    return Tensor._op(AddSequence, *variables, constant=constant)


def positive(a, where=True, constant=False):
    """ ``f(a) -> +a``

        Parameters
        ----------
        a : array_like

        where : numpy.ndarray
            Accepts a boolean array which is broadcast together
            with the operand(s). Values of True indicate to calculate
            the function at that position, values of False indicate
            to leave the value in the output alone.

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Positive, a, op_kwargs=(dict(where=where)), constant=constant)


def negative(a, where=True, constant=False):
    """ ``f(a) -> -a``

        Parameters
        ----------
        a : array_like

        where : numpy.ndarray
            Accepts a boolean array which is broadcast together
            with the operand(s). Values of True indicate to calculate
            the function at that position, values of False indicate
            to leave the value in the output alone.

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Negative, a, op_kwargs=(dict(where=where)), constant=constant)



import numpy as np
import mygrad as mg
from mygrad.nnet.losses import margin_ranking_loss

from numpy.testing import assert_allclose, assert_array_equal
import hypothesis.strategies as st
from hypothesis import given
import hypothesis.extra.numpy as hnp


def simple_loss(x1, x2, y, margin):
    """
    x1 : mygrad.Tensor, shape=(N, D)
    x2 : mygrad.Tensor, shape=(N, D)
    y : Union[int, numpy.ndarray], scalar or shape=(N,)
    margin : float

    Returns
    -------
    mygrad.Tensor, shape=()
    """
    y = np.asarray(y)
    if y.ndim:
        assert y.size == 1 or len(y) == len(x1)
        if x1.ndim == 2:
            y = y.reshape(-1, 1)

    return mg.mean(mg.maximum(0, margin - y * (x1 - x2)))


@given(shape=hnp.array_shapes(min_dims=1, max_dims=2),
       margin=st.floats(0, 1000),
       data=st.data())
def test_ranked_margin(shape, margin, data):
    x1 = data.draw(hnp.arrays(shape=shape, dtype=float, elements=st.floats(-1000, 1000)), label="x1")
    x2 = data.draw(hnp.arrays(shape=shape, dtype=float, elements=st.floats(-1000, 1000)), label="x2")
    y = data.draw(st.one_of(st.sampled_from((-1, 1)),
                            hnp.arrays(shape=shape[:1],
                                       dtype=int,
                                       elements=st.sampled_from((-1, 1)))),
                  label="y")

    x1_copy = np.copy(x1)
    x2_copy = np.copy(x2)
    y_copy = np.copy(y)

    x1_dum = mg.Tensor(x1)
    x2_dum = mg.Tensor(x2)

    x1_real = mg.Tensor(x1)
    x2_real = mg.Tensor(x2)

    loss_dum = simple_loss(x1_dum, x2_dum, y, margin)

    loss_real = margin_ranking_loss(x1_real, x2_real, y, margin)

    assert_allclose(actual=loss_real.data, desired=loss_dum.data, err_msg="losses don't match")

    assert_array_equal(x1, x1_copy, err_msg="`x1` was mutated by forward")
    assert_array_equal(x2, x2_copy, err_msg="`x2` was mutated by forward")
    if isinstance(y, np.ndarray):
        assert_array_equal(y, y_copy, err_msg="`y` was mutated by forward")

    loss_dum.backward()
    loss_real.backward()

    assert_allclose(actual=x1_real.grad, desired=x1_dum.grad, err_msg="x1.grad doesn't match")
    assert_allclose(actual=x2_real.grad, desired=x2_dum.grad, err_msg="x2.grad doesn't match")

    assert_array_equal(x1, x1_copy, err_msg="`x1` was mutated by backward")
    assert_array_equal(x2, x2_copy, err_msg="`x2` was mutated by backward")
    if isinstance(y, np.ndarray):
        assert_array_equal(y, y_copy, err_msg="`y` was mutated by backward")

    loss_real.null_gradients()
    assert x1_real.grad is None
    assert x2_real.grad is None



from mygrad.tensor_base import Tensor
from mygrad.nnet.layers import gru
from mygrad.nnet.activations import tanh, sigmoid
from mygrad import matmul

import hypothesis.strategies as st
from hypothesis import given, settings
import hypothesis.extra.numpy as hnp

import numpy as np
from numpy.testing import assert_allclose


def dense(x, y): return matmul(x, y)


@settings(deadline=None)
@given(st.data())
def test_gru_fwd(data):
    X = data.draw(hnp.arrays(shape=hnp.array_shapes(max_side=5, min_dims=3, max_dims=3),
                             dtype=float,
                             elements=st.floats(-10, 10)),
                  label="X")
    T, N, C = X.shape
    D = data.draw(st.sampled_from(list(range(1, 5))), label="D")
    dropout = data.draw(st.sampled_from([0, .45]), label="dropout")


    Wz = data.draw(hnp.arrays(shape=(D, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Wz")

    Uz = data.draw(hnp.arrays(shape=(C, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Uz")

    bz = data.draw(hnp.arrays(shape=(D,),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="bz")

    Wr = data.draw(hnp.arrays(shape=(D, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Wr")

    Ur = data.draw(hnp.arrays(shape=(C, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Ur")

    br = data.draw(hnp.arrays(shape=(D,),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="br")

    Wh = data.draw(hnp.arrays(shape=(D, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Wh")

    Uh = data.draw(hnp.arrays(shape=(C, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Uh")

    bh = data.draw(hnp.arrays(shape=(D,),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="bh")

    V = data.draw(hnp.arrays(shape=(D, C),
                             dtype=float,
                             elements=st.floats(-10.0, 10.0)),
                  label="V")


    s0 = np.zeros((N, D), dtype=float)

    X = Tensor(X)
    X2 = X.__copy__()

    Wz = Tensor(Wz)
    Wz2 = Wz.__copy__()

    Uz = Tensor(Uz)
    Uz2 = Uz.__copy__()

    bz = Tensor(bz)
    bz2 = bz.__copy__()

    Wr = Tensor(Wr)
    Wr2 = Wr.__copy__()

    Ur = Tensor(Ur)
    Ur2 = Ur.__copy__()

    br = Tensor(br)
    br2 = br.__copy__()

    Wh = Tensor(Wh)
    Wh2 = Wh.__copy__()

    Uh = Tensor(Uh)
    Uh2 = Uh.__copy__()

    bh = Tensor(bh)
    bh2 = bh.__copy__()

    V = Tensor(V)
    V2 = V.__copy__()

    s0 = Tensor(s0)
    s2 = s0.__copy__()

    s = gru(X, Uz, Wz, bz, Ur, Wr, br, Uh, Wh, bh, dropout=dropout, constant=True)
    o = dense(s[1:], V)
    ls = o.sum()

    assert s.constant is True

    if dropout:
        for d in [s.creator._dropUr, s.creator._dropUz, s.creator._dropUh,
                  s.creator._dropWr, s.creator._dropWz, s.creator._dropWh]:
            assert np.all(np.logical_or(d == 1 / (1 - dropout), d == 0))

    stt = s2
    all_s = [s0.data]
    ls2 = 0
    if dropout:
        Wz2d = s.creator._dropWz * Wz2
        Wr2d = s.creator._dropWr * Wr2
        Wh2d = s.creator._dropWh * Wh2
    else:
        Wz2d = Wz2
        Wr2d = Wr2
        Wh2d = Wh2
    for n, x in enumerate(X2):
        if not dropout:
            z = sigmoid(dense(x, Uz2) + dense(stt, Wz2d) + bz2)
            r = sigmoid(dense(x, Ur2) + dense(stt, Wr2d) + br2)
            h = tanh(dense(x, Uh2) + dense((r * stt), Wh2d) + bh2)
        else:
            z = sigmoid((s.creator._dropUz[0] * dense(x, Uz2)) + dense(stt, Wz2d) + bz2)
            r = sigmoid((s.creator._dropUr[0] * dense(x, Ur2)) + dense(stt, Wr2d) + br2)
            h =    tanh((s.creator._dropUh[0] * dense(x, Uh2)) + dense((r * stt), Wh2d) + bh2)

        stt = (1 - z) * h + z * stt
        all_s.append(stt)
        o = dense(stt, V2)
        ls2 += o.sum()

    tolerances = dict(atol=1e-5, rtol=1e-5)
    rec_s_dat = np.stack([i.data for i in all_s])

    assert_allclose(ls.data, ls2.data, **tolerances)

    assert_allclose(rec_s_dat, s.data, **tolerances)

    assert_allclose(Wz.data, Wz2.data, **tolerances)
    assert_allclose(Wr.data, Wr2.data, **tolerances)
    assert_allclose(Wh.data, Wh2.data, **tolerances)

    assert_allclose(Uz.data, Uz2.data, **tolerances)
    assert_allclose(Ur.data, Ur2.data, **tolerances)
    assert_allclose(Uh.data, Uh2.data, **tolerances)

    assert_allclose(bz.data, bz2.data, **tolerances)
    assert_allclose(br.data, br2.data, **tolerances)
    assert_allclose(bh.data, bh2.data, **tolerances)

    assert_allclose(V.data, V2.data, **tolerances)

    assert_allclose(X.data, X2.data, **tolerances)

    ls.null_gradients()
    for x in [s, Wz, Wr, Wh, bz, br, bh, X, Uz, Ur, Uh, V]:
        assert x.grad is None


@settings(deadline=None)
@given(st.data())
def test_gru_backward(data):
    tolerances = dict(atol=1e-5, rtol=1e-5)
    X = data.draw(hnp.arrays(shape=hnp.array_shapes(max_side=5, min_dims=3, max_dims=3),
                             dtype=float,
                             elements=st.floats(-10, 10)),
                  label="X")
    T, N, C = X.shape
    D = data.draw(st.sampled_from(list(range(1, 5))), label="D")
    dropout = data.draw(st.sampled_from([0, 0.45]), label="dropout")  # TODO: RESTORE DROPOUT


    Wz = data.draw(hnp.arrays(shape=(D, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Wz")

    Uz = data.draw(hnp.arrays(shape=(C, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Uz")

    bz = data.draw(hnp.arrays(shape=(D,),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="bz")

    Wr = data.draw(hnp.arrays(shape=(D, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Wr")

    Ur = data.draw(hnp.arrays(shape=(C, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Ur")

    br = data.draw(hnp.arrays(shape=(D,),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="br")

    Wh = data.draw(hnp.arrays(shape=(D, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Wh")

    Uh = data.draw(hnp.arrays(shape=(C, D),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="Uh")

    bh = data.draw(hnp.arrays(shape=(D,),
                              dtype=float,
                              elements=st.floats(-10.0, 10.0)),
                   label="bh")

    V = data.draw(hnp.arrays(shape=(D, C),
                             dtype=float,
                             elements=st.floats(-10.0, 10.0)),
                  label="V")

    s0 = np.zeros((N, D), dtype=float)

    X = Tensor(X)
    X2 = X.__copy__()

    Wz = Tensor(Wz)
    Wz2 = Wz.__copy__()

    Uz = Tensor(Uz)
    Uz2 = Uz.__copy__()

    bz = Tensor(bz)
    bz2 = bz.__copy__()

    Wr = Tensor(Wr)
    Wr2 = Wr.__copy__()

    Ur = Tensor(Ur)
    Ur2 = Ur.__copy__()

    br = Tensor(br)
    br2 = br.__copy__()

    Wh = Tensor(Wh)
    Wh2 = Wh.__copy__()

    Uh = Tensor(Uh)
    Uh2 = Uh.__copy__()

    bh = Tensor(bh)
    bh2 = bh.__copy__()

    V = Tensor(V)
    V2 = V.__copy__()

    s0 = Tensor(s0)
    s2 = s0.__copy__()

    s = gru(X, Uz, Wz, bz, Ur, Wr, br, Uh, Wh, bh, dropout=dropout, constant=False)
    o = dense(s[1:], V)
    ls = o.sum()
    ls.backward()

    stt = s2
    all_s = [s0.data]
    ls2 = 0
    if dropout:
        Wz2d = s.creator._dropWz * Wz2
        Wr2d = s.creator._dropWr * Wr2
        Wh2d = s.creator._dropWh * Wh2
    else:
        Wz2d = Wz2
        Wr2d = Wr2
        Wh2d = Wh2
    for n, x in enumerate(X2):
        if not dropout:
            z = sigmoid(dense(x, Uz2) + dense(stt, Wz2d) + bz2)
            r = sigmoid(dense(x, Ur2) + dense(stt, Wr2d) + br2)
            h = tanh(dense(x, Uh2) + dense((r * stt), Wh2d) + bh2)
        else:
            z = sigmoid((s.creator._dropUz[0] * dense(x, Uz2)) + dense(stt, Wz2d) + bz2)
            r = sigmoid((s.creator._dropUr[0] * dense(x, Ur2)) + dense(stt, Wr2d) + br2)
            h =    tanh((s.creator._dropUh[0] * dense(x, Uh2)) + dense((r * stt), Wh2d) + bh2)
        stt = (1 - z) * h + z * stt
        all_s.append(stt)
        o = dense(stt, V2)
        ls2 += o.sum()
    ls2.backward()

    rec_s_grad = np.stack([i.grad for i in all_s[1:]])

    assert_allclose(rec_s_grad, s.grad, **tolerances)

    assert_allclose(Wz.grad, Wz2.grad, **tolerances)
    assert_allclose(Wr.grad, Wr2.grad, **tolerances)
    assert_allclose(Wh.grad, Wh2.grad, **tolerances)

    assert_allclose(Uz.grad, Uz2.grad, **tolerances)
    assert_allclose(Ur.grad, Ur2.grad, **tolerances)
    assert_allclose(Uh.grad, Uh2.grad, **tolerances)

    assert_allclose(bz.grad, bz2.grad, **tolerances)
    assert_allclose(br.grad, br2.grad, **tolerances)
    assert_allclose(bh.grad, bh2.grad, **tolerances)

    assert_allclose(V.grad, V2.grad, **tolerances)

    assert_allclose(X.grad, X2.grad, **tolerances)

    ls.null_gradients()
    ls2.null_gradients()
    for x in [s, Wz, Wr, Wh, bz, br, bh, X, Uz, Ur, Uh, V]:
        assert x.grad is None



from mygrad import matmul

from tests.wrappers.uber import fwdprop_test_factory, backprop_test_factory
from tests.custom_strategies import broadcastable_shape

import numpy as np
from numpy.testing import assert_almost_equal

import hypothesis.strategies as st
from hypothesis import settings


@st.composite
def special_shape(draw, static_shape, shape=tuple(), min_dim=0, max_dim=5):
    """ search strategy that permits broadcastable dimensions to
        be prepended to a static shape - for the purposes of 
        drawing diverse shaped-arrays for matmul
        
        Returns
        -------
        hypothesis.searchstrategy.SearchStrategy
            -> Tuple[int, ...]"""
    return draw(broadcastable_shape(shape, min_dim, max_dim)) + static_shape


@fwdprop_test_factory(mygrad_func=matmul, true_func=np.matmul,
                      num_arrays=2, index_to_arr_shapes={0: (2, 3),
                                                         1: special_shape((3, 4))})
def test_matmul_fwd():
    """ a is n-d, b is n-d, a can broadcast into b"""
    pass

def test_matmul_fwd_static():
    a = [4., 3.]
    b = [1.3, 4.]
    assert_almost_equal(actual=matmul(a, b).data, desired=np.matmul(a, b))

    a = [[4., 3.], [1., 2.]]
    b = [4., 3.]
    assert_almost_equal(actual=matmul(a, b).data, desired=np.matmul(a, b))


@backprop_test_factory(mygrad_func=matmul, true_func=np.matmul, num_arrays=2,
                       index_to_arr_shapes={0: (4,), 1: (4,)}, as_decimal=False,
                       vary_each_element=True,
                       atol=1e-3, rtol=1e-3,
                       index_to_bnds={0: (-10, 10), 1: (-10, 10)})
def test_matmul_bkwd_1d_1d():
    """ a is 1-d, b is 1-d"""
    pass


@backprop_test_factory(mygrad_func=matmul, true_func=np.matmul, num_arrays=2, 
                       index_to_arr_shapes={0: special_shape((4,), min_dim=1, max_dim=2), 
                                            1: (4,)}, 
                       as_decimal=False,
                       vary_each_element=True,
                       atol=1e-3, rtol=1e-3,
                       index_to_bnds={0: (-10, 10), 1: (-10, 10)})
def test_matmul_bkwd_nd_1d():
    """ a is n-d, b is 1-d"""
    pass


@backprop_test_factory(mygrad_func=matmul, true_func=np.matmul, num_arrays=2, 
                       index_to_arr_shapes={1: special_shape((4, 1), min_dim=1, max_dim=2), 
                                            0: (4,)}, 
                       as_decimal=False,
                       vary_each_element=True,
                       atol=1e-3, rtol=1e-3,
                       index_to_bnds={0: (-10, 10), 1: (-10, 10)})
def test_matmul_bkwd_1d_nd():
    """ a is 1-d, b is n-d"""
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=matmul, true_func=np.matmul, num_arrays=2, 
                       index_to_arr_shapes={0: special_shape((4,), min_dim=1, max_dim=2), 
                                            1: (4, 5)}, 
                       as_decimal=False,
                       vary_each_element=True,
                       atol=1e-3, rtol=1e-3,
                       index_to_bnds={0: (-10, 10), 1: (-10, 10)})
def test_matmul_bkwd_nd_nd():
    """ a is n-d, b is n-d; b can broadcast into a"""
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=matmul, true_func=np.matmul, num_arrays=2, 
                       index_to_arr_shapes={0: (2, 4), 
                                            1: special_shape((4, 5), max_dim=2)}, 
                       as_decimal=False,
                       vary_each_element=True,
                       atol=1e-3, rtol=1e-3,
                       index_to_bnds={0: (-10, 10), 1: (-10, 10)})
def test_matmul_bkwd_nd_nd2():
    """ a is n-d, b is n-d; a can broadcast into b"""
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=matmul, true_func=np.matmul, num_arrays=2, 
                       index_to_arr_shapes={0: (2, 1, 3, 4), 
                                            1: (1, 2, 4, 2)}, 
                       as_decimal=False,
                       vary_each_element=True,
                       atol=1e-3, rtol=1e-3,
                       index_to_bnds={0: (-10, 10), 1: (-10, 10)})
def test_matmul_bkwd_nd_nd3():
    """ a is n-d, b is n-d; a and b broadcast mutually via singleton dimensions"""
    pass



from mygrad.tensor_base import Tensor
from mygrad.math.arithmetic.funcs import *
from mygrad.math.exp_log.funcs import *
from mygrad.math.trigonometric.funcs import *
from mygrad.math.hyperbolic_trig.funcs import *
from mygrad.math.sequential.funcs import *
from mygrad.math.sequential.funcs import max, min
from mygrad.math.nondifferentiable import argmax, argmin
from mygrad.math.misc.funcs import *
from mygrad.tensor_manip.array_shape.funcs import *
from mygrad.tensor_manip.transpose_like.funcs import *
from mygrad.tensor_creation.funcs import *
from mygrad.linalg.funcs import *

from mygrad.nnet.layers.utils import sliding_window_view

from ._version import get_versions

__version__ = get_versions()['version']
del get_versions


for attr in (sum, prod, cumprod, cumsum,
             mean, std, var,
             max, min,
             argmax, argmin,
             swapaxes, transpose, moveaxis,
             reshape, squeeze, ravel,
             matmul):
    setattr(Tensor, attr.__name__, attr)



"""
Tests Tensor/Operation implementation that uses topological sorting against the naive implementation.

Here, a rule-based state machine constructs computational graphs via 'rules' by which nodes (tensors)
are 'fused' via addition and multiplication operations. The state machine create and fuse nodes in
any patterns, invoke `null_gradients` and `clear_graph` arbitrarily as well.

The values and gradients of the nodes in the mygrad and naive graphs must match as an invariant to
any permutation of test states (i.e. permutations of the aforementioned rules)"""

import hypothesis.strategies as st
from hypothesis.stateful import Bundle, RuleBasedStateMachine, rule, invariant, precondition
from numpy.testing import assert_equal, assert_almost_equal
from mygrad import Tensor, add, multiply

from pytest import raises

from typing import List, Tuple

from .simple_graph import _add, _multiply, Node


def _node_ID_str(num): return "v{}".format(num + 1)


class GraphCompare(RuleBasedStateMachine):
    def __init__(self):
        super(GraphCompare, self).__init__()
        # stores the corresponding node/tensor v1, v2, ... as they are
        # created via the unit test (through `create_node` or `fuse_nodes`)
        # `Node` is the naive implementation of `Tensor` that we are checking
        # against
        self.node_list = []  # type: List[Tuple[Node, Tensor]]
        self.str_to_tensor_op = {"add": add, "multiply": multiply}
        self.str_to_node_op = {"add": _add, "multiply": _multiply}
        self.raised = False
    nodes = Bundle('nodes')

    @rule(target=nodes, value=st.floats(-10, 10), constant=st.booleans())
    def create_node(self, value, constant):
        n = Node(value, constant=constant)
        t = Tensor(value, constant=constant)
        self.node_list.append((n, t))
        return n, t

    @rule(target=nodes, a=nodes, b=nodes, op=st.sampled_from(["add", "multiply"]), constant=st.booleans())
    def fuse_nodes(self, a, b, op, constant):
        """
        Combine any pair of nodes (tensors) using either addition or multiplication, producing
        a new node (tensor)"""
        n_a, t_a = a
        n_b, t_b = b
        n_op = self.str_to_node_op[op]
        t_op = self.str_to_tensor_op[op]
        out = (n_op(n_a, n_b, constant=constant), t_op(t_a, t_b, constant=constant))
        self.node_list.append(out)
        return out
    
    @rule(items=nodes, clear_graph=st.booleans())
    def null_gradients(self, items, clear_graph):
        """
        Invoke `null_gradients` on the computational graph (naive and mygrad), with
        `clear_graph=True` specified optionally.
        """
        n, t = items
        n.null_gradients(clear_graph=clear_graph)
        t.null_gradients(clear_graph=clear_graph)

    @rule(items=nodes)
    def clear_graph(self, items):
        """
        Invoke `clear_graph` on the computational graph (naive and mygrad)
        """
        n, t = items
        n.clear_graph()
        t.clear_graph()

    @rule(items=nodes, grad=st.floats(-10, 10))
    def backprop(self, items, grad):
        """
        Invoke `backward(grad)` on the computational graph (naive and mygrad) from a randomly-selected
        node in the computational graph and using a randomly-generated gradient value.

        An exception should be raised if `clear_graph` is invoked anywhere prior to the invoking node.
        """
        n, t = items
        n.null_gradients(clear_graph=False)
        t.null_gradients(clear_graph=False)
        try:
            n.backward(grad, terminal_node=True)
        except Exception:
            with raises(Exception):
                t.backward(grad)
            self.raised = True
        else:
            t.backward(grad)

    @precondition(lambda self: not self.raised)
    @invariant()
    def all_agree(self):
        """
        Ensure that all corresponding nodes/tensors have matching data and gradients
        across the respective graphs.
        """
        for num, (n, t) in enumerate(self.node_list):
            assert bool(n._ops) is bool(t._ops), _node_ID_str(num)
            assert_equal(n.data, t.data, err_msg=_node_ID_str(num))
            if n.grad is None or t.grad is None:
                assert n.grad is t.grad, _node_ID_str(num)
            else:
                assert_almost_equal(desired=n.grad, actual=t.grad, err_msg=_node_ID_str(num))
            assert not t._accum_ops, _node_ID_str(num)


TestGraphComparison = GraphCompare.TestCase




from mygrad.tensor_base import Tensor
import numpy as np
from numpy.testing import assert_allclose

from ..wrappers.uber import fwdprop_test_factory, backprop_test_factory
from ..custom_strategies import basic_index, adv_integer_index
import hypothesis.extra.numpy as hnp

from hypothesis import settings

def test_getitem():
    x = Tensor([1, 2, 3])
    a, b, c = x
    f = 2*a + 3*b + 4*c
    f.backward()

    assert a.data == 1
    assert b.data == 2
    assert c.data == 3
    assert f.data == 20

    assert_allclose(a.grad, np.array(2))
    assert_allclose(b.grad, np.array(3))
    assert_allclose(c.grad, np.array(4))
    assert_allclose(x.grad, np.array([2, 3, 4]))


def get_item(*arrs, index, constant=False):
    o = arrs[0][index]
    if isinstance(o, Tensor):
        o._constant = constant
    return o


def basic_index_wrap(*arrs):
    return basic_index(arrs[0].shape)


def adv_index_int_wrap(*arrs):
    return adv_integer_index(arrs[0].shape)


def adv_index_bool_wrap(*arrs):
    return hnp.arrays(shape=arrs[0].shape, dtype=bool)


@fwdprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                      index_to_arr_shapes={0: hnp.array_shapes(max_side=6, max_dims=4)},
                      kwargs=dict(index=basic_index_wrap))
def test_getitem_basicindex_fwdprop():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                       index_to_arr_shapes={0: hnp.array_shapes(max_side=6, max_dims=4)},
                       kwargs=dict(index=basic_index_wrap), vary_each_element=True)
def test_getitem_basicindex_bkwdprop():
    pass


@fwdprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                      index_to_arr_shapes={0: hnp.array_shapes(max_side=6, max_dims=4)},
                      kwargs=dict(index=adv_index_int_wrap))
def test_getitem_advindex_int_fwdprop():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                       index_to_arr_shapes={0: hnp.array_shapes(max_side=6, max_dims=4)},
                       kwargs=dict(index=adv_index_int_wrap), vary_each_element=True)
def test_getitem_advindex_int_bkwdprop():
    pass


@fwdprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                      index_to_arr_shapes={0: hnp.array_shapes(max_side=6, max_dims=4)},
                      kwargs=dict(index=adv_index_bool_wrap))
def test_getitem_advindex_bool_fwdprop():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                       index_to_arr_shapes={0: hnp.array_shapes(max_side=6, max_dims=4)},
                       kwargs=dict(index=adv_index_bool_wrap), vary_each_element=True)
def test_getitem_advindex_bool_bkwdprop():
    pass


# test broadcast-compatible int-arrays
rows = np.array([0, 3], dtype=np.intp)
columns = np.array([0, 2], dtype=np.intp)


@fwdprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                      index_to_arr_shapes={0: (4, 3)},
                      kwargs=dict(index=np.ix_(rows, columns)))
def test_getitem_broadcast_index_fwdprop():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                       index_to_arr_shapes={0: (4, 3)},
                       kwargs=dict(index=np.ix_(rows, columns)), vary_each_element=True)
def test_getitem_broadcast_index_bkprop():
    pass


@fwdprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                      index_to_arr_shapes={0: (3, 2, 4, 3)},
                      kwargs=dict(index=(Ellipsis, 2, 0)))
def test_getitem_ellipsis_index_fwdprop():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                       index_to_arr_shapes={0: (3, 2, 4, 3)},
                       kwargs=dict(index=(Ellipsis, 2, 0)), vary_each_element=True)
def test_getitem_ellipsis_index_bkprop():
    pass


rows1 = np.array([False,  True, False,  True])
columns1 = [0, 2]


@fwdprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                      index_to_arr_shapes={0: (4, 3)},
                      kwargs=dict(index=np.ix_(rows1, columns1)))
def test_getitem_bool_int_fwdprop():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                       index_to_arr_shapes={0: (4, 3)},
                       kwargs=dict(index=np.ix_(rows1, columns1)), vary_each_element=True)
def test_getitem_bool_int_bkprop():
    pass


@fwdprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                      index_to_arr_shapes={0: (4, 3)},
                      kwargs=dict(index=(slice(1, 2), [1, 2])))
def test_getitem_basic_w_adv_fwdprop():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=get_item, true_func=get_item, num_arrays=1,
                       index_to_arr_shapes={0: (4, 3)},
                       kwargs=dict(index=(slice(1, 2), [1, 2])), vary_each_element=True)
def test_getitem_basic_w_adv_bkprop():
    pass



"""
Defines the base class for mathematical operations capable of back-propagating
gradients to their input tensors."""

from mygrad._utils import reduce_broadcast

import numpy as np

__all__ = ["Operation",
           "BroadcastableOp"]


class Operation:
    """ Base class for all tensor operations that support back-propagation
        of gradients.

        Consider the Operation-instance ``f``. A forward-pass through ``f`` is defined
        via ``f.__call__``. Thus, given tensors ``a`` and ``b``, a computational
        graph is defined ``f.__call__(a, b) -> c``, where the "creator" of tensor ``c``
        is recorded as ``f``::

              (node: a) --+
                           -> [operation: f(a, b)] --> (node: c)
              (node: b) --+

        Thus back-propagating through ``c`` will instruct ``f`` to back-propagate
        the gradient to its inputs, which are recorded as ``a`` and ``b``. Each
        node then back-propagates to any Operation-instance that is recorded
        as its creator, and so on.

        If an operation class has `scalar_only=True`, then the terminal node of a
        computational graph involving that operation can only trigger back-propagation
        from a 0-dimensional tensor (i.e. a scalar). This is `False` for operations that
        manifest as trivial element-wise operations over tensors. In such cases, the
        gradient of the operation can also be treated element-wise, and thus be computed
        unambiguously.
        """
    scalar_only = False

    def __call__(self, *input_vars):
        """ Performs a forward pass, f, of this Operation:
                  f(x1, ...., xn) -> out

            Parameters
            ----------
            *input_vars : mygrad.Tensor
                The input-arguments of f. The tuple (x1, ...., xn)
                should be bound to the instance-attribute `self.variables`

            Returns
            -------
            numpy.ndarray
                The output of the forward pass function."""

        self.variables = input_vars
        raise NotImplementedError

    def backward_var(self, grad, index, **kwargs):
        """ Given ``grad = d(out)/d(f)``, computes ``d(out)/d(var)``, and passes this result
            to ``var.backward()``, where var is the tensor-argument at position ``index``.

            Parameters
            ----------
            grad : numpy.ndarray
                The back-propagated total derivative with respect to the present
                operation (`f`): d(out)/df

            index : int
                The index-location of ``var`` in ``self.variables``

            Other Parameters
            ----------------
            _broadcastable : bool, optional (default:False)
                Devs-only: Indicates whether or not the up-stream operation
                can utilize broadcasting.

            Raises
            ------
            NotImplemented Error"""
        raise NotImplementedError

    def backward(self, grad, *, graph, **kwargs):
        """ Back-propagates the gradient through all of the operation's inputs.
        Constant tensors do not propagate a gradient.

        grad : numpy.ndarray
            The back-propagated total derivative with respect to the present
            operation (`f`): d(out)/df

        graph : Set[Operation]
            The set of all operations relevant to the terminal node of the computational graph,
            which triggered back-propagation.
        """
        for index, var in enumerate(self.variables):
            if not var.constant:
                if not var._ops:
                    raise Exception("Invalid Backprop: part of the computational graph containing "
                                    "this tensor was cleared prior to backprop")
                if var.grad is None:
                    tmp_grad = np.asarray(self.backward_var(grad, index, **kwargs))
                    var.grad = np.copy(tmp_grad) if np.shares_memory(tmp_grad, grad) else tmp_grad
                else:
                    var.grad += self.backward_var(grad, index, **kwargs)

        for var in {i for i in self.variables if not i.constant and i.creator is not None}:
            var._accum_ops.add(self)
            var._backward(graph=graph)


class BroadcastableOp(Operation):
    """ Signals that an Operation's forward pass can broadcast its tensor
        arguments."""
    def backward(self, grad, *, graph, **kwargs):
        """ Back-propagates the gradient through all of the operation's inputs.
        Constant tensors do not propagate a gradient.

        grad : numpy.ndarray
            The back-propagated total derivative with respect to the present
            operation (`f`): d(out)/df

        graph : Set[Operation]
            The set of all operations relevant to the terminal node of the computational graph,
            which triggered back-propagation.
        """
        for index, var in enumerate(self.variables):
            if not var.constant:
                if not var._ops:
                    raise Exception("Invalid Backprop: part of the computational graph containing "
                                    "this tensor was cleared prior to backprop")
                if var.grad is None:
                    tmp_grad = reduce_broadcast(self.backward_var(grad, index, **kwargs), var.shape)
                    var.grad = np.copy(tmp_grad) if np.shares_memory(tmp_grad, grad) else tmp_grad
                else:
                    var.grad += reduce_broadcast(self.backward_var(grad, index, **kwargs), var.shape)

        for var in {i for i in self.variables if not i.constant and i.creator is not None}:
            var._accum_ops.add(self)
            var._backward(graph=graph)



from tests.wrappers.uber import fwdprop_test_factory, backprop_test_factory
from mygrad import positive, negative, reciprocal, square
import numpy as np


def _is_non_zero(x):
    return np.all(np.abs(x.data) > 1e-3)


@fwdprop_test_factory(mygrad_func=positive, true_func=np.positive, num_arrays=1)
def test_positive_fwd():
    pass


@backprop_test_factory(mygrad_func=positive, true_func=np.positive, num_arrays=1)
def test_positive_backward():
    pass


@fwdprop_test_factory(mygrad_func=negative, true_func=np.negative, num_arrays=1)
def test_negative_fwd():
    pass


@backprop_test_factory(mygrad_func=negative, true_func=np.negative, num_arrays=1)
def test_negative_backward():
    pass


@fwdprop_test_factory(mygrad_func=reciprocal, true_func=np.reciprocal, num_arrays=1,
                      assumptions=_is_non_zero)
def test_reciprocal_fwd():
    pass


@backprop_test_factory(mygrad_func=reciprocal, true_func=np.reciprocal, num_arrays=1,
                       assumptions=_is_non_zero, atol=1e-5, rtol=1e-5)
def test_reciprocal_backward():
    pass


@fwdprop_test_factory(mygrad_func=square, true_func=np.square, num_arrays=1)
def test_square_fwd():
    pass


@backprop_test_factory(mygrad_func=square, true_func=np.square, num_arrays=1)
def test_square_backward():
    pass



def reduce_broadcast(grad, var_shape):
    """ Sum-reduce axes of `grad` so its shape matches `var_shape.

        This the appropriate mechanism for backpropagating a gradient
        through an operation in which broadcasting occurred for the
        given variable.

        Parameters
        ----------
        grad : numpy.ndarray
        var_shape : Tuple[int, ...]

        Returns
        -------
        numpy.ndarray

        Raises
        ------
        ValueError
            The dimensionality of the gradient cannot be less than
            that of its associated variable."""
    if grad.shape == var_shape:
        return grad

    if grad.ndim != len(var_shape):
        if grad.ndim < len(var_shape):
            raise ValueError("The dimensionality of the gradient of the broadcasted operation"
                             "is less than that of its associated variable")
        grad = grad.sum(axis=tuple(range(grad.ndim - len(var_shape))))

    keepdims = tuple(n for n, i in enumerate(grad.shape) if i != var_shape[n])
    if keepdims:
        grad = grad.sum(axis=keepdims, keepdims=True)

    return grad


""" Test `numerical_gradient`, `numerical_derivative`, and `broadcast_check`"""

from tests.utils.numerical_gradient import numerical_gradient, numerical_gradient_full

import hypothesis.extra.numpy as hnp
import hypothesis.strategies as st
from hypothesis import given

from tests.custom_strategies import valid_axes
import numpy as np
from numpy.testing import assert_allclose


def unary_func(x): return x ** 2


def binary_func(x, y): return x * y ** 2


def ternary_func(x, y, z): return z * x * y ** 2


@given(data=st.data(),
       x=hnp.arrays(shape=hnp.array_shapes(max_side=3, max_dims=3),
                    dtype=float,
                    elements=st.floats(-100, 100)),
       as_decimal=st.booleans())
def test_numerical_gradient_no_broadcast(data, x, as_decimal):
    atol, rtol = (1e-7, 1e-7) if as_decimal else (1e-2, 1e-2)
    y = data.draw(hnp.arrays(shape=x.shape,
                             dtype=float,
                             elements=st.floats(-100, 100)),
                  label="y")

    z = data.draw(hnp.arrays(shape=x.shape,
                             dtype=float,
                             elements=st.floats(-100, 100)),
                  label="z")

    grad = data.draw(hnp.arrays(shape=x.shape,
                                dtype=float,
                                elements=st.floats(-100, 100)),
                     label="grad")


    # check variable-selection
    assert numerical_gradient(unary_func, x, back_grad=grad,
                              vary_ind=[], as_decimal=as_decimal)[0] is None

    # no broadcast
    dx, = numerical_gradient(unary_func, x, back_grad=grad, as_decimal=as_decimal)

    assert_allclose(dx, grad * 2 * x, atol=atol, rtol=rtol)

    dx, dy = numerical_gradient(binary_func, x, y, back_grad=grad)
    assert_allclose(dx, grad * y ** 2, atol=atol, rtol=rtol)
    assert_allclose(dy, grad * 2 * x * y, atol=atol, rtol=rtol)

    dx, dy, dz = numerical_gradient(ternary_func, x, y, z, back_grad=grad)
    assert_allclose(dx, grad * z * y ** 2, atol=atol, rtol=rtol)
    assert_allclose(dy, grad * z * 2 * x * y, atol=atol, rtol=rtol)
    assert_allclose(dz, grad * x * y ** 2, atol=atol, rtol=rtol)


@given(data=st.data(),
       x=hnp.arrays(shape=(3, 4),
                    dtype=float,
                    elements=st.floats(-100, 100)),
       y=hnp.arrays(shape=(2, 3, 4),
                    dtype=float,
                    elements=st.floats(-100, 100)),
       grad=hnp.arrays(shape=(2, 3, 4),
                       dtype=float,
                       elements=st.floats(-100, 100)),
       as_decimal=st.booleans())
def test_numerical_gradient_x_broadcast(data, x, y, grad, as_decimal):
    atol, rtol = (1e-7, 1e-7) if as_decimal else (1e-2, 1e-2)

    # broadcast x
    dx, dy = numerical_gradient(binary_func, x, y, back_grad=grad)
    assert_allclose(dx, (grad * y ** 2).sum(axis=0), atol=atol, rtol=rtol)
    assert_allclose(dy, grad * 2 * x * y, atol=atol, rtol=rtol)


@given(data=st.data(),
       y=hnp.arrays(shape=(3, 4),
                    dtype=float,
                    elements=st.floats(-100, 100)),
       x=hnp.arrays(shape=(2, 3, 4),
                    dtype=float,
                    elements=st.floats(-100, 100)),
       grad=hnp.arrays(shape=(2, 3, 4),
                       dtype=float,
                       elements=st.floats(-100, 100)),
       as_decimal=st.booleans())
def test_numerical_gradient_y_broadcast(data, x, y, grad, as_decimal):
    atol, rtol = (1e-7, 1e-7) if as_decimal else (1e-2, 1e-2)

    # broadcast x
    dx, dy = numerical_gradient(binary_func, x, y, back_grad=grad)
    assert_allclose(dx, grad * y ** 2, atol=atol, rtol=rtol)
    assert_allclose(dy, (grad * 2 * x * y).sum(axis=0), atol=atol, rtol=rtol)


@given(data=st.data(),
       x=hnp.arrays(shape=(2, 1, 4),
                    dtype=float,
                    elements=st.floats(-100, 100)),
       y=hnp.arrays(shape=(1, 3, 4),
                    dtype=float,
                    elements=st.floats(-100, 100)),
       grad=hnp.arrays(shape=(2, 3, 4),
                       dtype=float,
                       elements=st.floats(-100, 100)),
       as_decimal=st.booleans())
def test_numerical_gradient_xy_broadcast(data, x, y, grad, as_decimal):
    atol, rtol = (1e-7, 1e-7) if as_decimal else (1e-2, 1e-2)

    # broadcast x
    dx, dy = numerical_gradient(binary_func, x, y, back_grad=grad)
    x_grad = (grad * y ** 2).sum(axis=1, keepdims=True)
    y_grad = (grad * 2 * x * y).sum(axis=0, keepdims=True)
    assert_allclose(dx, x_grad, atol=atol, rtol=rtol)
    assert_allclose(dy, y_grad, atol=atol, rtol=rtol)


@given(x=hnp.arrays(dtype=float,
                    elements=st.floats(-1, 1),
                    shape=(2,)),
       grad=hnp.arrays(dtype=float,
                       elements=st.floats(-1, 1),
                       shape=(2,)),
       as_decimal=st.booleans(),)
def test_numerical_gradient_vary_each(x, grad, as_decimal):
    atol, rtol = (1e-7, 1e-7) if as_decimal else (1e-2, 1e-2)
    dx, = numerical_gradient_full(lambda y: y[::-1], x, back_grad=np.array(grad),
                                  as_decimal=as_decimal)
    x_grad = grad[::-1]
    assert_allclose(actual=dx, desired=x_grad, atol=atol, rtol=rtol)



import mygrad as mg
from mygrad.tensor_base import Tensor
from mygrad.nnet.losses import softmax_crossentropy
from mygrad.nnet.activations import softmax
from mygrad import log

import numpy as np
from numpy.testing import assert_allclose
import hypothesis.strategies as st
from hypothesis import given
import hypothesis.extra.numpy as hnp

from pytest import raises

@given(st.data())
def test_softmax_crossentropy(data):
    """ Test the built-in implementation of multiclass hinge against the pure pygrad version"""
    s = data.draw(hnp.arrays(shape=hnp.array_shapes(max_side=10, min_dims=2, max_dims=2),
                             dtype=float,
                             elements=st.floats(-100, 100)))
    l = data.draw(hnp.arrays(shape=(s.shape[0],),
                             dtype=int,
                             elements=st.integers(min_value=0, max_value=s.shape[1] - 1)))
    scores = Tensor(s)
    softmax_cross = softmax_crossentropy(scores, l, constant=False)
    softmax_cross.backward()

    pygrad_scores = Tensor(s)
    probs = softmax(pygrad_scores)

    correct_labels = (range(len(l)), l)
    truth = np.zeros(pygrad_scores.shape)
    truth[correct_labels] = 1

    pygrad_cross = (-1/s.shape[0]) * (log(probs) * truth).sum()
    pygrad_cross.backward()
    assert_allclose(softmax_cross.data, pygrad_cross.data, atol=1e-5, rtol=1e-5)
    assert_allclose(scores.grad, pygrad_scores.grad, atol=1e-5, rtol=1e-5)


@given(shape=st.sampled_from([(3, 1), (3, 4), tuple()]))
def test_bad_label_shape(shape):
    """
    Ensures that softmax_crossentropy checks for shape-(N,) `y_true`
    """
    scores = mg.arange(12).reshape(3, 4)
    labels = mg.zeros(shape, dtype=int)
    with raises(ValueError):
        softmax_crossentropy(scores, labels)


@given(type=st.sampled_from([bool, float, np.float32]))
def test_bad_label_type(type):
    """
    Ensures that softmax_crossentropy checks integer-type `y_true`
    """
    scores = mg.arange(12).reshape(3, 4)
    labels = np.zeros((3,), dtype=type)
    with raises(TypeError):
        softmax_crossentropy(scores, labels)



from tests.wrappers.uber import fwdprop_test_factory, backprop_test_factory
from mygrad import cbrt
import numpy as np


def _is_non_zero(x):
    return np.all(np.abs(x.data) > 1e-5)


@fwdprop_test_factory(mygrad_func=cbrt, true_func=np.cbrt, num_arrays=1)
def test_cbrt_fwd():
    pass


@backprop_test_factory(mygrad_func=cbrt, true_func=np.cbrt,
                       index_to_bnds={0: (-100, 100)}, num_arrays=1,
                       assumptions=_is_non_zero,
                       as_decimal=False, atol=1e-5)
def test_cbrt_backward():
    pass



from tests.wrappers.uber import fwdprop_test_factory, backprop_test_factory
from mygrad import sqrt
import numpy as np


@fwdprop_test_factory(mygrad_func=sqrt, true_func=np.sqrt, num_arrays=1, index_to_bnds={0: (0, 100)})
def test_sqrt_fwd():
    pass


@backprop_test_factory(mygrad_func=sqrt, true_func=np.sqrt, num_arrays=1, index_to_bnds={0: (1e-5, 100)},
                       atol=1e-3)
def test_sqrt_backward():
    pass



from tests.wrappers.uber import fwdprop_test_factory, backprop_test_factory
from mygrad import sin, cos, tan, csc, sec, cot, sinc

import numpy as np


@fwdprop_test_factory(num_arrays=1, mygrad_func=sin, true_func=np.sin)
def test_sin_fwd():
    pass


@backprop_test_factory(num_arrays=1, mygrad_func=sin, true_func=np.sin, as_decimal=False,
                       atol=1e-6, rtol=1e-6)
def test_sin_backward():
    pass


@fwdprop_test_factory(num_arrays=1, mygrad_func=sinc, true_func=np.sinc)
def test_sinc_fwd():
    pass


@backprop_test_factory(num_arrays=1, mygrad_func=sinc, true_func=np.sinc, as_decimal=False,
                       atol=1e-5)
def test_sinc_backward():
    pass


@fwdprop_test_factory(num_arrays=1, mygrad_func=cos, true_func=np.cos)
def test_cos_fwd():
    pass


@backprop_test_factory(num_arrays=1, mygrad_func=cos, true_func=np.cos, as_decimal=False,
                       atol=1e-5)
def test_cos_backward():
    pass


@fwdprop_test_factory(num_arrays=1, mygrad_func=tan, true_func=np.tan,
                      index_to_bnds={0: (-np.pi/2 + 1e-5, np.pi/2 - 1e-5)})
def test_tan_fwd():
    pass


@backprop_test_factory(num_arrays=1, mygrad_func=tan, true_func=np.tan, as_decimal=False,
                       index_to_bnds={0: (-np.pi / 2 + 1e-5, np.pi / 2 - 1e-5)})
def test_tan_backward():
    pass


@fwdprop_test_factory(num_arrays=1, mygrad_func=csc, true_func=lambda x: 1 / np.sin(x),
                      index_to_bnds={0: (0 + 1e-5, np.pi - 1e-5)})
def test_csc_fwd():
    pass


@backprop_test_factory(num_arrays=1, mygrad_func=csc, true_func=lambda x: 1 / np.sin(x),
                       as_decimal=False,
                       index_to_bnds={0: (0 + 1e-5, np.pi - 1e-5)},
                       atol=1e-4, rtol=1e-4)
def test_csc_backward():
    pass


@fwdprop_test_factory(num_arrays=1, mygrad_func=sec, true_func=lambda x: 1 / np.cos(x),
                      index_to_bnds={0: (-np.pi/2 + 1e-5, np.pi/2 - 1e-5)})
def test_sec_fwd():
    pass


@backprop_test_factory(num_arrays=1, mygrad_func=sec, true_func=lambda x: 1 / np.cos(x),
                       index_to_bnds={0: (-np.pi/2 + 1e-5, np.pi/2 - 1e-5)}, atol=1e-4,
                       as_decimal=False)
def test_sec_backward():
    pass


@fwdprop_test_factory(num_arrays=1, mygrad_func=cot, true_func=lambda x: 1 / np.tan(x),
                      index_to_bnds={0: (0 + 1e-5, np.pi - 1e-5)})
def test_cot_fwd():
    pass


@backprop_test_factory(num_arrays=1, mygrad_func=cot, true_func=lambda x: 1 / np.tan(x),
                       index_to_bnds={0: (0 + 1e-5, np.pi - 1e-5)},
                       atol=1e-4, rtol=1e-4, as_decimal=False)
def test_cot_backward():
    pass



""" Test conv fwd-prop and back-prop for ND convs"""

from ..wrappers.uber import backprop_test_factory, fwdprop_test_factory
from ..utils.numerical_gradient import numerical_gradient_full

import hypothesis.extra.numpy as hnp
import hypothesis.strategies as st
from hypothesis import given, assume, settings

from pytest import raises

import numpy as np
from mygrad.nnet.layers import conv_nd
from mygrad import Tensor

from numpy.testing import assert_allclose


def get_outshape(x_shape, w_shape, stride, dilation):
    ''' Compute the shape of the output tensor given an input shape, convolutional
    filter shape, and stride.

    Parameters
    ----------
    x_shape : Tuple[int, ...]
        The shape of the input tensor.

    w_shape : Tuple[int, ...]
        The shape of the convolutional filter.

    stride : Tuple[int, ...]
        The stride at which to apply the convolutional filter to the input.

    dilation : Tuple[int, ...]
        The dilation used to form each window over the data.

    Returns
    -------
    numpy.ndarray[int], shape=(num_conv,)
        The shape of the output tensor resulting from convolving a tensor of shape `x_shape`
        with a tensor of shape `w_shape`.

        Returns `None` if an invalid combination of shapes are provided.
    '''
    x_shape = np.array(x_shape)
    w_shape = np.array(w_shape)
    stride = np.array(stride)
    dilation = np.array(dilation)
    out_shape = (x_shape - ((w_shape - 1) * dilation + 1)) / stride + 1

    if not all(i.is_integer() and i > 0 for i in out_shape):
        msg = "Stride and kernel dimensions are incompatible: \n"
        msg += "Input dimensions: {}\n".format(tuple(x_shape))
        msg += "Stride dimensions: {}\n".format(tuple(stride))
        msg += "Kernel dimensions: {}\n".format(tuple(w_shape))
        msg += "Dilation dimensions: {}\n".format(tuple(dilation))
        return None
    return out_shape.astype(np.int32)


def convolve_numpy(input_image, conv_filter, stride, dilation=None):
    ''' Convolve `input_image` with `conv_filter` at a stride of `stride`.

    Parameters
    ----------
    input_image : numpy.ndarray, shape=(C, H, ...)
        The input over which to perform convolution.

    conv_filter : numpy.ndarray, shape=(C, Hf, ...)
        The convolutional filter to slide across the image.

    stride : Sequence[int]
        The stride at which to apply `conv_filter` across `input_image`.

    Returns
    -------
    numpy.ndarray, shape=(H', ...)
        The result of convolving `input_image` with `conv_filter` at a stride of `stride`,
        where (H', W') is the result of `get_outshape`.
    '''
    conv_shape = conv_filter.shape[1:]
    in_shape = input_image.shape[1:]
    if dilation is None:
        dilation = (1,) * len(stride)
    out_shape = tuple(get_outshape(in_shape, conv_shape, stride, dilation))
    out = np.empty(out_shape, np.float32)
    for ind in np.ndindex(out_shape):
        slices = (slice(None),) + tuple(
            slice(i * s, i * s + w * d, d) for i, w, s, d in zip(ind, conv_shape, stride, dilation))
        out[ind] = np.sum(conv_filter * input_image[slices])
    return out


def conv_bank(input_images, conv_filters, stride, dilation=None):
    ''' Convolve a bank of filters over a stack of images.

    Parameters
    ----------
    input_images : numpy.ndarray, shape=(N, C, H, ...)
        The images over which to convolve our filters.

    conv_filters : numpy.ndarray, shape=(K, C, Hf, ...)
        The convolutional filters to apply to the images.

    stride : Sequence[int]
        The stride at which to apply each filter to the images.

    dilation : Sequence[int]

    Returns
    -------
    numpy.ndarray, shape=(N, K, H', ...)
        The result of convolving `input_image` with `conv_filter` at a stride of `stride`,
        where (H', ...) is the result of `get_outshape`.
    '''
    img_shape = input_images.shape[2:]
    conv_shape = conv_filters.shape[2:]
    if dilation is None:
        dilation = (1,) * len(stride)
    out_shape = get_outshape(img_shape, conv_shape, stride, dilation)

    out = np.empty((len(input_images), len(conv_filters), *out_shape))
    for i, image in enumerate(input_images):
        for j, conv in enumerate(conv_filters):
            out[i, j] = convolve_numpy(image, conv, stride, dilation)
    return out

## Defining Tests


def test_convnd_fwd_trivial():

    # trivial by-hand test: 1-dimensional conv
    # x:
    # [ 1,  2,  3,  4]

    # k:
    # [-1, -2],

    # stride = (2,)
    x = Tensor(np.arange(1, 5).reshape(1, 1, 4).astype(float))
    k = Tensor(-1 * np.arange(1, 3).reshape(1, 1, 2).astype(float))

    o = conv_nd(x, k, stride=(2,), constant=True)

    out = np.array([[[-5., -11.]]])
    assert isinstance(o, Tensor)
    assert o.constant is True
    assert o.scalar_only is False
    assert_allclose(actual=o.data, desired=out, err_msg="1d trivial test failed")


    # trivial by-hand test: 2-dimensional conv
    # x:
    # [ 1,  2,  3,  4],
    # [ 5,  6,  7,  8],
    # [ 9, 10, 11, 12]]

    # k:
    # [-1, -2],
    # [-3, -4]

    # stride = (1, 2)
    x = Tensor(np.arange(1, 13).reshape(1, 1, 3, 4).astype(float))
    k = Tensor(-1 * np.arange(1, 5).reshape(1, 1, 2, 2).astype(float))

    o = conv_nd(Tensor(x), k, stride=(1, 2), constant=True)

    out = np.array([[[[-44.,  -64.],
                      [-84., -104.]]]])
    assert isinstance(o, Tensor)
    assert o.constant is True
    assert o.scalar_only is False
    assert_allclose(actual=o.data, desired=out, err_msg="2d trivial test failed")


def test_bad_conv_shapes():
    x = np.zeros((1, 2, 2, 2))
    w = np.zeros((1, 3, 2, 2))
    with raises(AssertionError):
        conv_nd(x, w, stride=1, padding=0)  # mismatched channels

    w = np.zeros((1, 2, 3, 2))
    with raises(ValueError):
        conv_nd(x, w, stride=1, padding=0)  # large filter

    w = np.zeros((1, 2, 2, 2))
    with raises(AssertionError):
        conv_nd(x, w, stride=0, padding=0)  # bad stride

    with raises(AssertionError):
        conv_nd(x, w, stride=[1, 2, 3])  # bad stride

    with raises(AssertionError):
        conv_nd(x, w, stride=1, padding=-1)  # bad pad

    with raises(AssertionError):
        conv_nd(x, w, stride=1, padding=[1, 2, 3])  # bad pad

    with raises(ValueError):
        conv_nd(x, w, stride=3, padding=1)  # shape mismatch


@settings(deadline=None)
@given(data=st.data(),
       x=hnp.arrays(dtype=float, shape=hnp.array_shapes(max_dims=6, min_dims=3, max_side=6),
                    elements=st.floats(-10, 10)),
       num_filters=st.integers(1, 3))
def test_conv_ND_fwd(data, x, num_filters):
    """ Test convs 1D-4D with various strides and dilations."""
    x = x[0:min(x.shape[0], 3), 0:min(x.shape[1], 3)]
    win_dim = x.ndim - 2
    win_shape = data.draw(st.tuples(*(st.integers(1, s) for s in x.shape[-win_dim:])), label="win_shape")
    kernels = data.draw(hnp.arrays(dtype=float, shape=(num_filters, x.shape[1], *win_shape),
                                   elements=st.floats(-10, 10)),
                        label="kernels")

    stride = data.draw(st.tuples(*(st.integers(1, s) for s in x.shape[-win_dim:])), label="stride")

    max_dilation = np.array(x.shape[-win_dim:]) // win_shape
    dilation = data.draw(st.tuples(*(st.integers(1, s) for s in max_dilation)), label="dilation")
    conf = dict(stride=stride, dilation=dilation)

    # skip invalid data/kernel/stride/dilation combinations
    assume(get_outshape(x.shape[2:], kernels.shape[2:], stride, dilation) is not None)
    numpy_conv = conv_bank(x, kernels, **conf)
    mygrad_conv = conv_nd(x, kernels, **conf).data
    assert_allclose(actual=mygrad_conv, desired=numpy_conv, atol=1e-6, rtol=1e-6)


@fwdprop_test_factory(mygrad_func=conv_nd, true_func=conv_bank, num_arrays=2,
                      index_to_arr_shapes={0: (4, 5, 7), 1: (2, 5, 3)},
                      kwargs=dict(stride=(1,), dilation=(1,)))
def test_conv_1d_fwd():
    """ (N=4, C=5, W=7) x (F=2, C=5, Wf=3); stride=1, dilation=1

    Also tests meta properties of conv function - appropriate return type,
    behavior with `constant` arg, etc."""
    pass


def _conv_nd(x, w, stride, dilation=1):
    """ use mygrad-conv_nd forward pass for numerical derivative

        Returns
        -------
        numpy.ndarray"""
    return conv_nd(x, w, stride=stride, dilation=dilation, constant=True).data


settings(deadline=None)
@given(data=st.data(),
       x=hnp.arrays(dtype=float, shape=hnp.array_shapes(max_dims=5, min_dims=3, max_side=6),
                    elements=st.floats(-10, 10)),
       num_filters=st.integers(1, 3))
def test_conv_ND_bkwd(data, x, num_filters):
    """ Test conv-backprop 1D-3D with various strides and dilations."""
    x = x[0:min(x.shape[0], 1), 0:min(x.shape[1], 1)]
    win_dim = x.ndim - 2
    win_shape = data.draw(st.tuples(*(st.integers(1, s) for s in x.shape[-win_dim:])), label="win_shape")
    kernels = data.draw(hnp.arrays(dtype=float, shape=(num_filters, x.shape[1], *win_shape),
                                   elements=st.floats(-10, 10)),
                        label="kernels")

    stride = data.draw(st.tuples(*(st.integers(1, s) for s in x.shape[-win_dim:])), label="stride")

    max_dilation = np.array(x.shape[-win_dim:]) // win_shape
    dilation = data.draw(st.tuples(*(st.integers(1, s) for s in max_dilation)), label="dilation")
    conf = dict(stride=stride, dilation=dilation)

    # skip invalid data/kernel/stride/dilation combinations
    assume(get_outshape(x.shape[2:], kernels.shape[2:], stride, dilation) is not None)

    x = Tensor(x)
    kernels = Tensor(kernels)

    out = conv_nd(x, kernels, **conf)
    grad = data.draw(hnp.arrays(shape=out.shape,
                                dtype=float,
                                elements=st.floats(-10, 10),
                                unique=True),
                     label="grad")

    out.backward(grad)
    grads_numerical = numerical_gradient_full(_conv_nd, *(i.data for i in (x, kernels)),
                                              back_grad=grad, kwargs=conf,
                                              as_decimal=False)

    for n, (arr, d_num) in enumerate(zip((x, kernels), grads_numerical)):
        assert_allclose(arr.grad, d_num, atol=1e-4, rtol=1e-4,
                        err_msg="arr-{}: numerical derivative and mygrad derivative do not match".format(n))


@backprop_test_factory(mygrad_func=conv_nd,
                       true_func=_conv_nd,
                       num_arrays=2,
                       index_to_arr_shapes={0: (2, 1, 7), 1: (2, 1, 3)},
                       kwargs={"stride": (1,)},
                       index_to_bnds={0: (-10, 10), 1: (-10, 10)},
                       vary_each_element=True, as_decimal=False, atol=1e-4, rtol=1e-4)
def test_conv_1d_bkwd():
    """ (N=2, C=1, W=7) x (F=2, C=1, Wf=3); stride=1, dilation=1

    Also tests meta properties of conv-backprop - appropriate return type,
    behavior with `constant` arg, good behavior of null_gradients, etc."""
    pass




from mygrad.tensor_base import Tensor
from mygrad import matmul
from mygrad.nnet.layers import simple_RNN
from mygrad.nnet.activations import tanh


import hypothesis.strategies as st
from hypothesis import given, settings
import hypothesis.extra.numpy as hnp

import numpy as np
from numpy.testing import assert_allclose


@settings(deadline=None)
@given(st.data())
def test_recurrent(data):
    X = data.draw(hnp.arrays(shape=hnp.array_shapes(max_side=5, min_dims=3, max_dims=3),
                             dtype=float,
                             elements=st.floats(-10, 10)))
    T, N, C = X.shape
    D = data.draw(st.sampled_from(list(range(1, 5))))

    s0 = data.draw(hnp.arrays(shape=(N, D),
                              dtype=float,
                              elements=st.floats(0.0, 0.0)))

    W = data.draw(hnp.arrays(shape=(D, D),
                             dtype=float,
                             elements=st.floats(-10.0, 10.0)))

    U = data.draw(hnp.arrays(shape=(C, D),
                             dtype=float,
                             elements=st.floats(-10.0, 10.0)))

    V = data.draw(hnp.arrays(shape=(D, C),
                             dtype=float,
                             elements=st.floats(-10.0, 10.0)))

    X = Tensor(X)
    X2 = X.__copy__()

    W = Tensor(W)
    W2 = W.__copy__()

    U = Tensor(U)
    U2 = U.__copy__()

    V = Tensor(V)
    V2 = V.__copy__()

    s0 = Tensor(s0)
    s2 = s0.__copy__()

    s = simple_RNN(X, U, W)
    o = matmul(s[1:], V)
    ls = o.sum()
    ls.backward()

    stt = s2
    all_s = [s0.data]
    ls2 = 0
    for n, x in enumerate(X2):
        stt = tanh(matmul(x, U2) + matmul(stt, W2))
        all_s.append(stt)
        o = matmul(stt, V2)
        ls2 += o.sum()
    ls2.backward()

    rec_s_dat = np.stack([i.data for i in all_s])
    rec_s_grad = np.stack([i.grad for i in all_s[1:]])
    assert_allclose(rec_s_dat, s.data, atol=1e-5, rtol=1e-5)
    assert_allclose(rec_s_grad, s.grad[1:], atol=1e-5, rtol=1e-5)
    assert_allclose(ls.data, ls2.data, atol=1e-5, rtol=1e-5)
    assert_allclose(W.data, W2.data, atol=1e-5, rtol=1e-5)
    assert_allclose(W.grad, W2.grad, atol=1e-5, rtol=1e-5)
    assert_allclose(U.data, U2.data, atol=1e-5, rtol=1e-5)
    assert_allclose(U.grad, U2.grad, atol=1e-5, rtol=1e-5)
    assert_allclose(V.data, V2.data, atol=1e-5, rtol=1e-5)
    assert_allclose(V.grad, V2.grad, atol=1e-5, rtol=1e-5)
    assert_allclose(X.data, X2.data, atol=1e-5, rtol=1e-5)
    assert_allclose(X.grad, X2.grad, atol=1e-5, rtol=1e-5)

    ls.null_gradients()
    for x in [s, W, X, U, V]:
        assert x.grad is None



from mygrad._utils import reduce_broadcast
from tests.custom_strategies import broadcastable_shape
import numpy as np

from numpy.testing import assert_allclose

from pytest import raises
from hypothesis import given, assume
import hypothesis.strategies as st
import hypothesis.extra.numpy as hnp


def test_bad_gradient_dimensionality():
    """ test that grad.dim < len(var_shape) raises ValueError"""
    var_shape = (1, 2, 3)
    grad = np.empty((1, 2))
    with raises(ValueError):
        reduce_broadcast(grad=grad, var_shape=var_shape)


@given(grad=hnp.arrays(dtype=float, shape=hnp.array_shapes(), elements=st.floats(-100, 100)))
def test_broadcast_scalar(grad):
    """ test when grad was broadcasted from a scalar"""
    assert_allclose(reduce_broadcast(grad, tuple()), grad.sum())


@given(grad=hnp.arrays(dtype=float, shape=hnp.array_shapes(), elements=st.floats(-100, 100)))
def test_reduce_broadcast_same_shape(grad):
    """ test when no broadcasting occurred"""
    var_shape=grad.shape
    reduced_grad = reduce_broadcast(grad=grad, var_shape=var_shape)
    assert_allclose(actual=reduced_grad, desired=grad)


@given(var_shape=hnp.array_shapes(min_side=2),
       data=st.data())
def test_reduce_broadcast_nokeepdim(var_shape, data):
    """ example broadcasting: (2, 3) -> (5, 2, 3)"""
    grad_shape = data.draw(broadcastable_shape(shape=var_shape,
                                               min_dim=len(var_shape) + 1,
                                               max_dim=len(var_shape) + 3),
                           label="grad_shape")
    assume(1 not in grad_shape[-len(var_shape):])
    grad = np.ones(grad_shape, dtype=float)

    reduced_grad = reduce_broadcast(grad=grad, var_shape=var_shape)
    reduced_grad *= np.prod(var_shape) / grad.size  # scale reduced-grad so all elements are 1
    assert_allclose(actual=reduced_grad, desired=np.ones(var_shape))


@given(var_shape=hnp.array_shapes(),
       data=st.data())
def test_reduce_broadcast_keepdim(var_shape, data):
    """ example broadcasting: (2, 1, 4) -> (2, 5, 4)"""
    grad = data.draw(hnp.arrays(dtype=float,
                     shape=broadcastable_shape(shape=var_shape,
                                               min_dim=len(var_shape),
                                               max_dim=len(var_shape)),
                                               elements=st.just(1.)),
                     label='grad')

    reduced_grad = reduce_broadcast(grad=grad, var_shape=var_shape)
    assert reduced_grad.shape == tuple(i if i < j else j for i, j in zip(var_shape, grad.shape))
    assert (i == 1 for i, j in zip(var_shape, grad.shape) if i < j)
    sum_axes = tuple(n for n, (i, j) in enumerate(zip(var_shape, grad.shape)) if i != j)
    assert_allclose(actual=reduced_grad, desired=grad.sum(axis=sum_axes, keepdims=True))


@given(grad=hnp.arrays(dtype=float, shape=(5, 3, 4, 2), elements=st.floats(-.01, .01)))
def test_hybrid_broadcasting(grad):
    """ tests new-dim and keep-dim broadcasting
         (3, 1, 2) -> (5, 3, 4, 2)"""
    var_shape = (3, 1, 2)
    reduced = reduce_broadcast(grad=grad, var_shape=var_shape)
    answer = grad.sum(axis=0).sum(axis=-2, keepdims=True)
    assert_allclose(actual=reduced, desired=answer)





